import datetime
import os
import yaml
import json

from azure_cis_scanner import utils
from azure_cis_scanner.utils import load_resource_groups

activity_logs_path = os.path.join(config['raw_data_dir'], 'activity_logs.json')
storage_accounts_path = os.path.join(config['raw_data_dir'], 'storage_accounts.json')
resource_groups_path = os.path.join(config['raw_data_dir'], 'resource_groups.json')
storage_accounts_filtered_path = os.path.join(config['filtered_data_dir'], 'storage_accounts_filtered.json')

def get_storage_accounts(storage_accounts_path):
    """
    Query Azure api for storage accounts info and save to disk
    """
    storage_accounts_cmd = "az storage account list"
    storage_accounts = json.loads(utils.call(storage_accounts_cmd))
        
    with open(storage_accounts_path, 'w') as f:
        json.dump(storage_accounts, f, indent=4, sort_keys=True)
    return storage_accounts

def load_storage_accounts(storage_accounts_path):
    with open(storage_accounts_path, 'r') as f:
        storage_accounts = yaml.safe_load(f)
    return storage_accounts

# timedelta=90 days can cause BadRequest error depending on timezone
# https://github.com/Azure/azure-cli/issues/4885
activity_log_timedelta_days = 89
activity_logs_starttime_timedelta = datetime.timedelta(days=activity_log_timedelta_days)
def get_start_time(timedelta=activity_logs_starttime_timedelta):
    """
    Given datetime.timedelta(days=days, hours=hours), return string in iso tz format 
    """
    return datetime.datetime.strftime(datetime.datetime.now() - timedelta, "%Y-%m-%dT%H:%M:%SZ")

def get_activity_logs(activity_logs_path, resource_groups):
    if os.path.exists(activity_logs_path):
        print("activity_logs_path {} exists, using existing values".format(activity_logs_path))
        return
    activity_logs = {}
    start_time = get_start_time(activity_logs_starttime_timedelta)
    for resource_group in resource_groups:
        resource_group = resource_group['name']
        activity_log_cmd = "az monitor activity-log list --resource-group {resource_group} --start-time {start_time}".format(
            resource_group=resource_group, start_time=start_time)
        activity_log = json.loads(utils.call(activity_log_cmd))
        activity_logs[resource_group] = activity_log
    with open(activity_logs_path, 'w') as f:
        json.dump(activity_logs, f, indent=4, sort_keys=True)
    return activity_logs    

def load_activity_logs(activity_logs_path):
    with open(activity_logs_path, 'r') as f:
        activity_logs = yaml.safe_load(f)
    return activity_logs



#################
# Tests
#################

def secure_transfer_required_is_set_to_enabled_3_1(storage_accounts):
    items_flagged_list = []
    for account in storage_accounts:
        name = account['name']
        resource_group = account['resourceGroup']
        enabled = account['enableHttpsTrafficOnly']
        if enabled != True:
            items_flagged_list.append((resource_group, name))
    stats = {'items_flagged': len(items_flagged_list), "items_checked": len(storage_accounts)}
    metadata = {"finding_name": "secure_transfer_required_is_set_to_enabled",
                "negative_name": "secure_transfer_required_not_enabled",
                "columns": ["Resource Group", "Storage Account Name"]}
    return {"items": items_flagged_list, 
            "stats": stats, 
            "metadata": metadata }
            

def storage_service_encryption_is_set_to_enabled_for_blob_service_3_2(storage_accounts):
    items_flagged_list = []
    for account in storage_accounts:
        if account['encryption']['services']['blob'] and (account['encryption']['services']['blob']['enabled'] != True):
            items_flagged_list.append((account['resourceGroup'], account['name']))

    stats = {'items_flagged': len(items_flagged_list),
             'items_checked': len(storage_accounts)}
    metadata = {"finding_name": "storage_service_encryption_is_set_to_enabled_for_blob_service",
                "negative_name": "storage_service_encryption_not_enabled_for_blob_service",
                "columns": ["Resource Group","Storage Account Name"]}
    
    return {"items": items_flagged_list, "stats": stats, "metadata": metadata }
           

# may need to run section 6 Networking first to get activity_log
def storage_account_access_keys_are_periodically_regenerated_3_3(activity_logs, storage_accounts, resource_groups):
    items_flagged_list = []
    
    max_rotation_days = 90
    most_recent_rotations = {}
    for resource_group in resource_groups:
        resource_group_name = resource_group['name']
        for log in activity_logs[resource_group_name]:
            if log["authorization"] and (log["authorization"]["action"] == "Microsoft.Storage/storageAccounts/regenerateKey/action"):
                scope = log["authorization"]["scope"]
                _, _, _, resource_group, _, _, _, storage_account_name = scope.split('/')
                timestamp = log["eventTimestamp"]
                event_day = timestamp.split('T')[0]
                event_day = datetime.datetime.strptime(event_time, '%Y-%m-%d')
                status = log["status"]["localizedValue"]
                if status == "Success":
                    # fromtimestamp(0) gives smallest date possible in epoch time
                    existing_update = most_recent_rotations.get(storage_account, datetime.datetime.fromtimestamp(0))
                    most_recent_rotations[storage_account] = max(existing_update, event_time)

    for storage_account in storage_accounts:
        resource_group = storage_account["resourceGroup"]
        storage_account_name = storage_account['name']
        items_flagged_list.append((resource_group, storage_account_name, str(most_recent_rotations.get(storage_account_name, "No rotation"))))

    stats = {'items_flagged': len(items_flagged_list),
             'items_checked': len(storage_accounts)}
    metadata = {"finding_name": "storage_account_access_keys_are_periodically_regenerated",
                "negative_name": "storage_account_access_keys_not_periodically_regenerated",
                "columns": ["Resource Group", "Storage Account", "Rotation Date"]}
    
    return {"items": items_flagged_list, "stats": stats, "metadata": metadata}

def shared_access_signature_tokens_expire_within_an_hour_3_4(storage_accounts):
    """
    There is no automation possible for this currently
    Manual
    """
    pass

def shared_access_signature_tokens_are_allowed_only_over_https_3_5(storage_accounts):
    """
    There is no automation possible for this currently
    Manual
    """
    pass
                                      
def storage_service_encryption_is_set_to_enabled_for_file_service_3_6(storage_accounts):
    items_flagged_list = []
    stats = {}
    for account in storage_accounts:
        if account['encryption']['services']['file'] and (account['encryption']['services']['file']['enabled'] != True):
            items_flagged_list.append((account['name']))

    stats = {'items_flagged': len(items_flagged_list),
             'items_checked': len(storage_accounts)}
    metadata = {"finding_name": "storage_service_encryption_is_set_to_enabled_for_file_service",
                "negative_name": "storage_service_encryption_not_enabled_for_file_service",
                "columns": ["Storage Account Name"]}

    return {"items": items_flagged_list, "stats": stats, "metadata": metadata}

def public_access_level_is_set_to_private_for_blob_containers_3_7(storage_accounts):
    items_flagged_list = []
    items_checked = 0
    for account in storage_accounts:
        account_name = account["name"]
        resource_group = account["resourceGroup"]
        # get a key that works.  likely this will be a specific key not key[0]
        keys_cmd = "az storage account keys list --account-name {account_name} --resource-group {resource_group}".format(
            account_name=account_name, resource_group=resource_group)
        keys = json.loads(utils.call(keys_cmd))
        key = keys[0]
        container_list_cmd = "az storage container list --account-name {account_name} --account-key {account_key}".format(
            account_name=account_name, account_key=account_key)
        container_list = json.loads(utils.call(container_list_cmd))
        for container in container_list:
            print(container)
            items_checked += 1
            public_access = container["properties"]["public_access"]
            if public_access == True:
                items_flagged_list.append((account_name, container))
    stats = {'items_flagged': len(items_flagged_list), "items_checked": items_checked}
    metadata = {"finding_name": "public_access_level_is_set_to_private_for_blob_containers",
                "negative_name": "public_access_level_not_private_for_blob_containers",
                "columns": ["Storage Account Name", "Container"]}
    
    return {"items": items_flagged_list, "stats": stats, "metadata": metadata }

    
def get_data():
    """
    Generate json for the storage_accounts findings
    """
    resource_groups = load_resource_groups(resource_groups_path)
    print(resource_groups)
    get_activity_logs(activity_logs_path, resource_groups)
    get_storage_accounts(storage_accounts_path)

def test_controls():
    """
    Generate filtered (failing) output in json
    """
    resource_groups = load_resource_groups(resource_groups_path)
    storage_accounts = load_storage_accounts(storage_accounts_path)
    activity_logs = load_activity_logs(activity_logs_path)
    
    storage_results = {}
    storage_results['secure_transfer_required_is_set_to_enabled'] = secure_transfer_required_is_set_to_enabled_3_1(storage_accounts)
    storage_results['storage_service_encryption_is_set_to_enabled_for_blob_service'] = storage_service_encryption_is_set_to_enabled_for_blob_service_3_2(storage_accounts)
    storage_results['storage_account_access_keys_are_periodically_regenerated'] = storage_account_access_keys_are_periodically_regenerated_3_3(activity_logs, storage_accounts, resource_groups)
    storage_results['storage_service_encryption_is_set_to_enabled_for_file_service'] = storage_service_encryption_is_set_to_enabled_for_file_service_3_6(storage_accounts)
    #storage_results['public_access_level_is_set_to_private_for_blob_containers'] = public_access_level_is_set_to_private_for_blob_containers_3_7(storage_accounts)
        
    with open(storage_accounts_filtered_path, 'w') as f:
        json.dump(storage_results, f, indent=4, sort_keys=True)
    return storage_results



from flask import Flask, render_template, redirect, url_for, session, make_response, render_template_string
from flask_nav import Nav
from flask_nav.elements import *
import argparse
import os
import numpy as np
import pandas as pd
import io
import yaml
import re 
import datetime
import random

from azure_cis_scanner import utils
import settings
from render_utils import *

HAS_MATPLOTLIB = False

try:
    from matplotlib.backends.backend_agg import FigureCanvasAgg as FigureCanvas
    from matplotlib.figure import Figure
    from matplotlib.dates import DateFormatter
    import matplotlib.pyplot as plt
    HAS_MATPLOTLIB = True
except ImportError as e:
    print(e)
    print("Unable to import matplotlib.  No graphing available.")


app = Flask(__name__)
nav = Nav(app)
app.secret_key = os.urandom(16)
app.config['SESSION_TYPE'] = 'filesystem'

@app.route('/')
def index_base():
    return redirect("/{}".format(app.config['ACTIVE_SUBSCRIPTION_DIR']), code=302)

@app.route('/_subscription_dir')
def _subscription_dir():
    selected_active_subscription_dir = request.args.get('state', session.get('active_subscription_dir', None))
    print('selected_active_subscription_dir2', selected_active_subscription_dir)

    print('subscription_dir route', subscription_dir)

    session['ACTIVE_SUBSCRIPTION_DIR'] = subscription_dir
    return jsonify({'selected': subscription_dir})

@app.route('/<active_subscription_dir>')
def index(active_subscription_dir, methods=['POST','GET']):
    selected_active_subscription_dir = request.args.get('selected', session.get('ACTIVE_SUBSCRIPTION_DIR', active_subscription_dir))
    session['active_subscription_dir'] = selected_active_subscription_dir
    state = {'selected', selected_active_subscription_dir}
    print('selected_active_subscription_dir', selected_active_subscription_dir)
    if selected_active_subscription_dir != active_subscription_dir:
        print('redirecting', selected_active_subscription_dir)
        return redirect('/'+selected_active_subscription_dir)
    else:
        subscription_dirs = [(subscription_dir_from_account(account), subscription_dir_from_account(account)) for account in accounts]
        print('subscription_dirs', subscription_dirs)
        return render_template('index.html', 
            active_subscription_dir=active_subscription_dir, 
            state=state,
            subscription_dirs=subscription_dirs)

@app.route('/services/<service>')
def service(service):
    findings_table = [(x['subsection_number'], x['subsection_name'], get_finding_name(x['finding_name'], x['subsection_name'])) for x in cis_structure['TOC'][service]]
    stats = get_latest_stats(app.config['SCANS_DATA_DIR'])
    #pprint.pprint("service:stats: {}".format(stats))
    return render_template('service.html', service=service, title=title_except(service), findings_table=findings_table, stats=stats)

@app.route('/services/<service>/<finding>')
def finding(service, finding):
    """
    Render the non-graph portion of the finding as a table of the latest date recording this finding
    The graph portion is rendered in plot_finding
    """
    finding_entry = get_finding_index(cis_structure['TOC'][service], finding)
    service_data = get_filtered_data_by_name(app.config['SCANS_DATA_DIR'], service)
    pprint.pprint("finding:service_data: {}".format(service_data))
    error_str = ''
    if not service_data:
        error_str = 'No section named "{}" found\n'.format(service)
    finding_name = get_finding_name(finding_entry['finding_name'], finding_entry['subsection_name'])
    finding_data = service_data.get(finding_name, None)
    print("finding_data {}".format(finding_data))
    if not finding_data:
        error_str += 'No finding named "{}" in {}\n'.format(finding, service_data.keys())
    elif (not finding_data.get("metadata", None)):
        error_str += 'No metadata for "{}"\n'.format(finding)
    elif not finding_data.get("metadata").get("columns", None):
        error_str += 'No columns in metadata for "{}"\n'.format(finding)
    elif not finding_data.get("items", None):
        error_str += 'No items list for "{}"\n'.format(finding)
    elif not finding_data.get("stats", None):
        error_str += 'No stats section in finding_data {}'.format(finding)
    elif not finding_data.get("stats", None):
        error_str += 'No stats section in finding_data {}'.format(finding)
    if error_str:
        return render_template('finding.html', service=service, finding=finding,
            finding_entry=finding_entry, table='', title=title_except(finding), error_str=error_str, items_checked=0)
    else:
        items_checked=finding_data['stats']['items_checked']
        data = pd.DataFrame(finding_data['items'], columns=finding_data['metadata']['columns'])
        return render_template('finding.html', service=service, finding=finding,
            finding_entry=finding_entry, table=data.to_html(), title=title_except(finding), items_checked=items_checked)

@app.route("/subscription_dir/<subscription_dir>")
def set_subscription_dir(subscription_dir):
    active_subscription_dir = subscription_dir
    return redirect("/{}".format(active_subscription_dir), code=302)


def subscription_dir_from_account(account):
    return account['name'].split(' ')[0] + '-' + account['id'].split('-')[0]

# Todo: use wrappers to generate the nav bars.  Too tricky for now.  Generated with print statements.
# def build_navbar(cis_structure):
#     navbar = [('Home', 'index')]
#     kwarg_list = [ dict(service=section) for section in cis_structure['section_ordering'] ]
#     arg_list = [ (section, 'section') for section in cis_structure['section_ordering'] ]
#     navbar = list(map(mappable_f, zip(arg_list, kwarg_list)))
#     print(navbar)
#     return navbar

@app.route("/graphs/<service>/<finding>.png")
def plot_finding(service, finding):
    if not HAS_MATPLOTLIB:
        return make_response("Install matplotlib for graphing support")
    section_name = service
    subsection_name = finding
    stats = get_stats(app.config['SCANS_DATA_DIR'])
    df = multi_index_df_from_stats(stats)
    finding_df = df.loc[section_name].loc[subsection_name]
    if len(finding_df) < 2:
        return make_response('Need at least 2 days of data for a plot.  Come back tomorrow!')
    finding_df = finding_df.fillna(method='ffill')
    y = finding_df["Flagged"].tolist()
    y = np.array(y)
    x = finding_df.index
    print("plot_finding x: ", x)
    print("plot_finding y: ", y)
    mask = np.isfinite(y)

    fig, ax = plt.subplots()
    ax.plot_date(x, y, '-')
    ax.xaxis.set_major_formatter(DateFormatter('%Y-%m-%d'))
    fig.autofmt_xdate()
    ax.axes.set_ylabel('Items Flagged for Finding')
    ax.axes.set_xlabel('Date')
    ax.set_title(subsection_name, verticalalignment='bottom')
    line, = ax.plot(x[mask],y[mask], ls="--",lw=1)
    ax.plot(x,y, color=line.get_color(), lw=1.5)

    canvas=FigureCanvas(fig)
    png_output = io.BytesIO()
    canvas.print_png(png_output)
    response=make_response(png_output.getvalue())
    response.headers['Content-Type'] = 'image/png'
    return response

def multi_index_tuples_from_stats(stats):
    """
    Create a multi-index panel from stats
    
    The major axis is a multi-index is created from the first two levels of a dict (section, finding)
    The minor axis is the date
    The data columns are items_checked, items_flagged
    """
    frames = []
    tuples = []
    index_tuples = []
    for section in cis_structure['section_ordering']:
        for subsection in cis_structure['TOC'][section]:
            subsection_name = subsection['subsection_name']
            finding_name = '_'.join(map(str.lower, subsection_name.split(' ')))
            start_date, end_date = min_max_dates()
            print("start_date {} end_date {}".format(start_date, end_date))
            for date in sorted(pd.date_range(start=start_date, end=end_date, freq='D')):
                str_date = datetime_to_dir_date(date)
                has_data = True
                if section not in stats:
                    has_data = False
                elif finding_name not in stats[section]:
                    has_data = False
                if has_data and (str_date in stats[section][finding_name]):
                    item_stats = stats[section][finding_name][str_date]
                    items_checked, items_flagged = item_stats['items_checked'], item_stats['items_flagged']
                else:
                    items_checked, items_flagged = None, None
                tuples.append((section, subsection_name, date, items_flagged, items_checked))
                index_tuples.append((section, subsection_name, date))                
    return tuples, index_tuples
            
def multi_index_df_from_stats(stats):
    tuples, index_tuples = multi_index_tuples_from_stats(stats)
    multi_index = pd.MultiIndex.from_tuples(index_tuples, names=["section", "finding", "date"])
    print(multi_index)
    return pd.DataFrame(tuples, index=multi_index, columns=["section_drop", "finding_drop", "date","Flagged", "Checked"])

def dir_date_to_datetime(string):
    """
    Takes format of directory '%Y-%m-%d' and converts to datetime"""
    return datetime.datetime.strptime(string, "%Y-%m-%d")

def datetime_to_dir_date(dt):
    """
    Takes format of directory '%Y-%m-%d' and converts to datetime"""
    return dt.strftime("%Y-%m-%d")

def min_max_dates():
    dates = list(map(dir_date_to_datetime, get_dirs(app.config['SCANS_DATA_DIR'])))
    return min(dates), max(dates)


nav.register_element(
    'top',
    Navbar(
        View('Identity and Access Management', 'service', service='Identity and Access Management'),
        View('Security Center', 'service', service='Security Center'),
        View('Storage Accounts', 'service', service='Storage Accounts'),
        View('SQL Servers', 'service', service='SQL Servers'),
        View('SQL Databases', 'service', service='SQL Databases'),
        View('Logging and Monitoring', 'service', service='Logging and Monitoring'),
        View('Networking', 'service', service='Networking'),
        View('Virtual Machines', 'service', service='Virtual Machines'),
        View('Other Security Considerations', 'service', service='Other Security Considerations'),
        #Text(active_subscription_dir)
        #Subgroup(active_subscription_dir, *[Link(account['name']) for account in accounts], [])
        )
    )

def main():
    mainparser = argparse.ArgumentParser()
    mainparser.add_argument('--tenant-id', default=None, help='azure tenant id, if None, use default.  Scanner assumes different runs/project dirs for distinct tenants')
    mainparser.add_argument('--subscription-id', default=None, help='azure subscription id, if None, use default, if "all" use all subscriptions with default tenant')
    mainparser.add_argument('--use-api-for-auth', default=True, help='if false, use azure cli calling subprocess, else use python-azure-sdk')
    # TODO, set default in __init__.py or somewhere and make it windows compatible
    mainparser.add_argument('--scans-dir', default='/engagements/cis_test', help='base dir of where to place or load files')

    parser = mainparser.parse_args()

    # TODO figure out better way to get base dir or let user select in UI
    credentials_tuples = utils.set_credentials_tuples(parser)

    tenant_id, subscription_id, subscription_name, credentials = credentials_tuples[0]
    subscription_dirname = utils.get_subscription_dirname(subscription_id, subscription_name)
    active_subscription_dir = subscription_dirname
    print("dir(utils)", dir(utils))
    scans_dir = utils.set_scans_dir(parser.scans_dir)
    
    app.config['ACTIVE_SUBSCRIPTION_DIR'] = active_subscription_dir
    app.config['SCANS_DIR'] = scans_dir
    app.config['SCANS_DATA_DIR'] = os.path.join(scans_dir, active_subscription_dir)

    app.run(debug=True, host='0.0.0.0')

# Might need something like https://testdriven.io/developing-a-single-page-app-with-flask-and-vuejs for menu drop-down multi-subscription.

# def create_app(configfile=None):
#     app = Flask(__name__)
#     nav.init_app(app)

#     # not good style, but like to keep our examples short
#     @app.route('/')
#     def index():
#         return render_template('index.html')

#     @app.route('/services/<service>')
#     def service(service):
#         findings_table = [(x['section_number'], x['section_name']) for x in service['findings'].items()]
#         print(findings_table)
#         return render_template('service.html', service=service, findings_table=findings_table)

#     @app.route('/services/<service>/<finding>')
#     def finding(service, finding):
#         return render_template('finding.html', service=service, finding=finding)

#     return app


if __name__ == "__main__":
    main()





import json
import os
import yaml

from azure.common.client_factory import get_client_from_cli_profile, get_client_from_auth_file
from azure.mgmt.compute import ComputeManagementClient
from azure.mgmt.resource import ResourceManagementClient, SubscriptionClient
from msrestazure.azure_active_directory import MSIAuthentication

from azure_cis_scanner import utils

filtered_virtual_machines_path = os.path.join(config['filtered_data_dir'], 'virtual_machines_filtered.json')
virtual_machines_path = os.path.join(config['raw_data_dir'], 'virtual_machines.json')
disks_path = os.path.join(config['raw_data_dir'], 'disks.json')
snapshots_path = os.path.join(config['raw_data_dir'], 'snapshots.json')
resource_groups_path = os.path.join(config['raw_data_dir'], "resource_groups.json")
credentials = config['cli_credentials']
sp_credentials = config['sp_credentials']
subscription_id=config['subscription_id']

def get_vms(sp_credentials, subscription_id=subscription_id):
    virtual_machines = []
    resource_groups = utils.load_resource_groups(resource_groups_path)
    compute = ComputeManagementClient(credentials, subscription_id)
    for resource_group in [x['name'] for x in resource_groups]:
        vms = compute.virtual_machines.list(resource_group)
        virtual_machines.extend(utils.get_list_from_paged_results(vms))
    with open(virtual_machines_path, 'w') as f:
        json.dump(virtual_machines, f, indent=4, sort_keys=True)
    return virtual_machines

def get_disks(disks_path):
    compute = ComputeManagementClient(sp_credentials, subscription_id)
    disks = compute.disks.list()
    
    disks = utils.get_list_from_paged_results(disks)
    with open(disks_path, 'w') as f:
        json.dump(disks, f, indent=4, sort_keys=True)
    return disks

def get_snapshots(snapshots_path):
    compute = ComputeManagementClient(sp_credentials, subscription_id)
    snaps = compute.snapshots.list()
    snaps = utils.get_list_from_paged_results(snaps)

    with open(snapshots_path, 'w') as f:
        json.dump(snaps, f, indent=4, sort_keys=True)
    return snaps

def get_virtual_machines(virtual_machines_path):
    """
    @virtual_machines_path: string - path to output json file
    @returns: list of virtual_machines dict
    """
    virtual_machines = json.loads(utils.call("az vm list"))
    with open(virtual_machines_path, 'w') as f:
        json.dump(virtual_machines, f, indent=4, sort_keys=True)
    return virtual_machines

def load_virtual_machines(virtual_machines_path):
    with open(virtual_machines_path, 'r') as f:
        virtual_machines = yaml.load(f)
    return virtual_machines

def load_disks(disks_path):
    with open(disks_path, 'r') as f:
        disks = yaml.load(f)
    return disks

def load_snapshots(snapshots_path):
    with open(snapshots_path, 'r') as f:
        snaps = yaml.load(f)
    return snaps

def get_data():
    get_virtual_machines(virtual_machines_path)
    # TODO switch to api below, but the datastructure is less detailed.  Trace cli call.
    #get_vms(virtual_machines_path)
    get_snapshots(snapshots_path)
    get_disks(disks_path)

def vm_agent_is_installed_7_1(virtual_machines):
    items_flagged_list = []
    for vm in virtual_machines:
        has_agent = False
        if vm['resources']:
            for resource in vm["resources"]:
                if ((vm['resources'][0]['virtualMachineExtensionType'] == 'MicrosoftMonitoringAgent') and (vm['resources'][0]['provisioningState'] == 'Succeeded')):
                    has_agent = True
        if has_agent:
            items_flagged_list.append((vm['resourceGroup'], vm['name']))
    
    stats = {'items_flagged': len(items_flagged_list),
             'items_checked': len(virtual_machines)}
    metadata = {"finding_name": "vm_agent_is_installed",
                "negative_name": "",
                "columns": ["Resource Group", "Name"]}            
    return  {"items": items_flagged_list, "stats": stats, "metadata": metadata}

def os_disk_is_encrypted_7_2(virtual_machines):
    items_flagged_list = []
    items_checked = 0
    for vm in virtual_machines:
        if vm['storageProfile']['osDisk']['encryptionSettings']:
            if not (vm['storageProfile']['osDisk']['encryptionSettings']['enabled'] == True):
                items_flagged_list.append((vm['resourceGroup'], vm['name'], vm['storageProfile']['osDisk']['name']))
                items_checked += 1
        else:
            items_flagged_list.append((vm['resourceGroup'], vm['name'], vm['storageProfile']['osDisk']['name']))
            items_checked += 1

    stats = {'items_flagged': len(items_flagged_list),
             'items_checked': len(virtual_machines)}
    metadata = {"finding_name": "os_disk_is_encrypted",
                "negative_name": "",
                "columns": ["Resource Group", "Name", "Disk Name"]}            
    return  {"items": items_flagged_list, "stats": stats, "metadata": metadata}

def data_disks_are_encrypted_7_3(virtual_machines, disks, snapshots):
    items_flagged_list = []
    items_checked = 0
    for vm in virtual_machines:
        name = vm['name']
        resource_group = vm['resourceGroup']
        encrypted = utils.call("az vm encryption show --name {name} --resource-group {resource_group} --query dataDisk".format(
            name=name, resource_group=resource_group))
        if encrypted in ["", "Azure Disk Encryption is not enabled"]:
            items_flagged_list.append((vm['resourceGroup'], vm['name'], "unknown"))
    for disk in disks:
        if ('encryptionSettings' not in disk) or \
        (disk['encryptionSettings'] == None) or \
        (not disk['encryptionSettings']['enabled']):
            items_flagged_list.append((disk['location'], "unknown", disk['name']))

    # There doesn't seem to be encryption info in the snapshot.json
    # for snap in snapshots:
    #     if snap['encryptionSettings'] == None:
    #         items_flagged_list.append((disk['resourceGroup'], "unknown", disk['name']))


    stats = {'items_flagged': len(items_flagged_list), 
             'items_checked': len(virtual_machines)}
    metadata = {"finding_name": "data_disks_are_encrypted",
                "negative_name": "",
                "columns": ["Resource Group", "Name", "Disk Name"]}
    return  {"items": items_flagged_list, "stats": stats, "metadata": metadata}


def only_approved_extensions_are_installed_7_4(virtual_machines):
    # items in the following list do not imply failure, but require review
    items_flagged_list = []
    approved_extensions = [
        'AzureDiskEncryption',
        'IaaSAntimalware',
        'IaaSDiagnostics',
        'MicrosoftMonitoringAgent',
        'SqlIaaSAgent',
        'OmsAgentForLinux', 
        'VMAccessForLinux',
    ]
    for vm in virtual_machines:
        name = vm['name']
        resource_group = vm['resourceGroup']
        extensions = json.loads(utils.call("az vm extension list --vm-name {name} --resource-group {resource_group}".format(name=name,
            resource_group=resource_group)))
        for extension in extensions:
            if extension['virtualMachineExtensionType'] not in approved_extensions:
                items_flagged_list.append((resource_group, name, extension['virtualMachineExtensionType']))
    
    stats = {'items_flagged': len(items_flagged_list),
             'items_checked': len(virtual_machines)}
    metadata = {"finding_name": "only_approved_extensions_are_installed",
                "negative_name": "",
                "columns": ["Resource Group", "VM Name", "Extension Name"]}            
    return  {"items": items_flagged_list, "stats": stats, "metadata": metadata}

def latest_patches_for_all_virtual_machines_are_applied_7_5(virtual_machines):
    pass

def endpoint_protection_for_all_virtual_machines_is_installed_7_6(virtual_machines):
    items_flagged_list = []
    accepted_protections = set(['EndpointSecurity', 'TrendMicroDSA', 'Antimalware', 'EndpointProtection','SCWPAgent', 'PortalProtectExtension', 'FileSecurity', 'IaaSAntimalware'])
    for vm in virtual_machines:
        name = vm['name']
        resource_group = vm['resourceGroup']
        extensions = json.loads(utils.call("az vm extension list --vm-name {name} --resource-group {resource_group}".format(name=name,
            resource_group=resource_group)))
        has_protection = False
        for extension in extensions:
            if set([extension['virtualMachineExtensionType']]).intersection(accepted_protections):
                has_protection = True
        if not has_protection:
            items_flagged_list.append((resource_group, name))

    stats = {'items_flagged': len(items_flagged_list),
             'items_checked': len(virtual_machines)}
    metadata = {"finding_name": "endpoint_protection_for_all_virtual_machines_is_installed",
                "negative_name": "",
                "columns": ["Resource Group", "Name"]}            
    return  {"items": items_flagged_list, "stats": stats, "metadata": metadata}


def test_controls():
    results = {}
    virtual_machines = load_virtual_machines(virtual_machines_path)
    disks = load_disks(disks_path)
    snapshots = load_snapshots(snapshots_path)
    results['vm_agent_is_installed'] = vm_agent_is_installed_7_1(virtual_machines)
    results['os_disk_is_encrypted'] = os_disk_is_encrypted_7_2(virtual_machines)
    results['data_disks_are_encrypted'] = data_disks_are_encrypted_7_3(virtual_machines, disks, snapshots)
    results['only_approved_extensions_are_installed'] = only_approved_extensions_are_installed_7_4(virtual_machines)
    results['endpoint_protection_for_all_virtual_machines_is_installed'] = endpoint_protection_for_all_virtual_machines_is_installed_7_6(virtual_machines)
    
    with open(filtered_virtual_machines_path, 'w') as f:
        json.dump(results, f, indent=4, sort_keys=True)
    return results


#%load {scanner_dir}/scanner/modules/logging_and_monitoring.py

# Generate files in raw_data_dir

import datetime
import json
import os
import traceback
import yaml

from azure.mgmt.monitor import MonitorManagementClient
from azure_cis_scanner import utils

from azure_cis_scanner.utils import get_list_from_paged_results, get_service_principal_credentials, AzScannerException

monitor_diagnostic_settings_path = os.path.join(config['raw_data_dir'], 'monitor_diagnostic_settings.json')
monitor_log_profiles_path = os.path.join(config['raw_data_dir'], 'monitor_log_profiles.json')
activity_logs_path = os.path.join(config['raw_data_dir'], 'activity_logs.json')
activity_log_alerts_path = os.path.join(config['raw_data_dir'], 'activity_log_alerts.json')
resource_groups_path = os.path.join(config['raw_data_dir'], "resource_groups.json")
resource_ids_for_diagnostic_settings_path = os.path.join(config['raw_data_dir'], 'resource_ids_for_diagnostic_settings.json')
resource_diagnostic_settings_path = os.path.join(config['raw_data_dir'], 'resource_diagnostic_settings.json')
logging_and_monitoring_filtered_path = os.path.join(config['filtered_data_dir'], 'logging_and_monitoring_filtered.json')
credentials = config['cli_credentials']
subscription_id = config['subscription_id']

def get_resource_ids_for_diagnostic_settings():
    resource_ids = []
    # Other resource_ids could be gathered.  So far, only keyvault
    keyvaults = json.loads(utils.call("az keyvault list"))
    for keyvault in keyvaults:
        resource_ids.append(keyvault['id'])
    with open(resource_ids_for_diagnostic_settings_path, 'w') as f:
        json.dump(resource_ids, f, indent=4, sort_keys=True)
    return resource_ids

def load_resource_ids_for_diagnostic_settings(resource_ids_for_diagnostic_settings_path):
    with open(resource_ids_for_diagnostic_settings_path, 'r') as f:
        resource_ids_for_diagnostic_settings = yaml.load(f)
    return resource_ids_for_diagnostic_settings

def get_resource_diagnostic_settings(resource_ids_for_diagnostic_settings):
    keyvault_settings_list = []
    for resource_id in resource_ids_for_diagnostic_settings:
        keyvault_settings = json.loads(utils.call("az monitor diagnostic-settings list --resource {resource_id}".format(resource_id=resource_id)))
        *prefix, resource_group, _, _, _, keyvault_name = resource_id.split('/')
        if not keyvault_settings['value']:
            keyvault_settings['value'].append({'keyvault_name': keyvault_name, 'resourceGroup': resource_group})
        else:
            for setting in keyvault_settings['value']:
                setting['keyvault_name'] = keyvault_name
        print(keyvault_settings)
        keyvault_settings_list.append(keyvault_settings)
        
    with open(resource_diagnostic_settings_path, 'w') as f:
        yaml.dump(keyvault_settings_list, f)
    return resource_ids_for_diagnostic_settings 

def load_resource_diagnostic_settings(resource_diagnostic_settings_path):
    with open(resource_diagnostic_settings_path, 'r') as f:
        resource_diagnostic_settings = yaml.load(f)
    return resource_diagnostic_settings        
        
def get_monitor_diagnostic_settings(monitor_diagnostic_settings_path, resource_ids):
    """
    @monitor_diagnostic_settings_path: string - path to output json file
    @returns: list of activity_log_alerts dicts
    """
    monitor_diagnostic_settings_results = {}
    for resource_id in resource_ids:
        monitor_diagnostic_settings = json.loads(utils.call("az monitor diagnostic-settings list --resource {resource_id}".format(resource_id=resource_id)))
        monitor_diagnostic_settings_results[resource_id] = monitor_diagnostic_settings
    with open(monitor_diagnostic_settings_path, 'w') as f:
        json.dump(monitor_diagnostic_settings_results, f, indent=4, sort_keys=True)
    return monitor_diagnostic_settings_results

def load_monitor_diagnostic_settings(monitor_diagnostic_settings_path):
    with open(monitor_diagnostic_settings_path, 'r') as f:
        monitor_diagnostic_settings = yaml.load(f)
    return monitor_diagnostic_settings


def get_monitor_log_profiles(monitor_log_profiles_path):
    monitor_log_profiles = json.loads(utils.call("az monitor log-profiles list"))
    with open(monitor_log_profiles_path, 'w') as f:
        json.dump(monitor_log_profiles, f, indent=4, sort_keys=True)
    return monitor_log_profiles

def load_monitor_log_profiles(monitor_log_profiles_path):
    with open(monitor_log_profiles_path, 'r') as f:
        monitor_log_profiles = yaml.load(f)
    return monitor_log_profiles

# duplicated in storage_accounts.log
# timedelta=90 days can cause BadRequest error depending on timezone
# https://github.com/Azure/azure-cli/issues/4885
activity_log_timedelta_days = 89
activity_logs_starttime_timedelta = datetime.timedelta(days=activity_log_timedelta_days)
def get_start_time(timedelta=activity_logs_starttime_timedelta):
    """
    Given datetime.timedelta(days=days, hours=hours), return string in iso tz format 
    """
    return datetime.datetime.strftime(datetime.datetime.now() - timedelta, "%Y-%m-%dT%H:%M:%SZ")

def get_activity_logs(activity_logs_path, resource_groups):
    if os.path.exists(activity_logs_path):
        print("activity_logs_path {} exists, using existing values".format(activity_logs_path))
        return
    activity_logs = {}
    start_time = get_start_time(activity_logs_starttime_timedelta)
    for resource_group in resource_groups:
        resource_group = resource_group['name']
        activity_log = json.loads(utils.call("az monitor activity-log list --resource-group {resource_group} --start-time {start_time}".format(
            resource_group=resource_group, start_time=start_time)))
        activity_logs[resource_group] = activity_log
    with open(activity_logs_path, 'w') as f:
        json.dump(activity_logs, f, indent=4, sort_keys=True)
    return activity_logs    

def load_activity_logs(activity_logs_path):
    with open(activity_logs_path, 'r') as f:
        activity_log = yaml.load(f)
    return activity_log

def get_activity_log_alerts(activity_log_alerts_path):
    activity_log_alerts = json.loads(utils.call("az monitor activity-log alert list"))
    with open(activity_log_alerts_path, 'w') as f:
        json.dump(activity_log_alerts, f, indent=4, sort_keys=True)
    return activity_log_alerts   

def load_activity_log_alerts(activity_log_alerts_path):
    with open(activity_log_alerts_path, 'r') as f:
        activity_log_alerts = yaml.load(f)
    return activity_log_alerts

def get_data():
    resource_ids_for_diagnostic_settings = get_resource_ids_for_diagnostic_settings()
    resource_groups = utils.load_resource_groups(resource_groups_path)
    get_monitor_log_profiles(monitor_log_profiles_path)
    get_monitor_diagnostic_settings(monitor_diagnostic_settings_path, resource_ids_for_diagnostic_settings)
    get_activity_log_alerts(activity_log_alerts_path)
    get_activity_logs(activity_logs_path, resource_groups)
    get_resource_diagnostic_settings(resource_ids_for_diagnostic_settings)

    
    
##################
# Tests
##################

def test_controls():
    """
    Use the data in raw_data_dir or in memory to run tests.
    Filtered output of raw_data_dir for failing systems is placed in filtered_data_dir
    """
    resource_ids_for_diagnostic_settings = load_resource_ids_for_diagnostic_settings(resource_ids_for_diagnostic_settings_path)
    resource_diagnostic_settings = load_resource_diagnostic_settings(resource_diagnostic_settings_path)
    resource_groups = utils.load_resource_groups(resource_groups_path)
    monitor_log_profiles = load_monitor_log_profiles(monitor_log_profiles_path)
    monitor_diagnostic_settings = load_monitor_diagnostic_settings(monitor_diagnostic_settings_path)
    activity_log_alerts = load_activity_log_alerts(activity_log_alerts_path)
    activity_logs = load_activity_logs(activity_logs_path)
    
    
    results = {}
    results['a_log_profile_exists'] = a_log_profile_exists_5_1(monitor_log_profiles)
    results['activity_log_retention_is_set_365_days_or_greater'] = activity_log_retention_is_set_365_days_or_greater_5_2(monitor_log_profiles)
    results['activity_log_alert_is_configured'] = activity_log_alert_is_configured(activity_log_alerts, log_alert_policies)
    results['logging_for_azure_keyvault_is_enabled'] = logging_for_azure_keyvault_is_enabled_5_13(resource_diagnostic_settings)

    with open(logging_and_monitoring_filtered_path, 'w') as f:
        json.dump(results, f, indent=4, sort_keys=True)
    return results


def a_log_profile_exists_5_1(monitor_log_profiles):
    items_flagged_list = []
    if monitor_log_profiles:
        pass
    else:
        items_flagged_list.append(("No log profile"))
        
    stats = {'items_flagged': len(items_flagged_list),
             'items_checked': min(1, len(items_flagged_list))}
    metadata = {"finding_name": "a_log_profile_exists",
                "negative_name": "",
                "columns": ["Monitor Log Profile"]}            
    return  {"items": items_flagged_list, "stats": stats, "metadata": metadata}
    

# Todo, untested as we have [] for log-profiles
#@gen_results(results)
def activity_log_retention_is_set_365_days_or_greater_5_2(monitor_log_profiles):
    items_flagged_list = []
    if monitor_log_profiles:
        for monitor_log_profile in monitor_log_profiles:
            days = monitor_log_profile.get('retentionPolicy', {}).get('days', -1)
            if monitor_log_profile.get('retentionPolicy', {}).get('days') <= MIN_ACTIVITY_LOG_RETENDION_DAYS:
                items_flagged_list.append((monitor_log_profile['id'], monitor_log_profile['storageAccountId'], days))
    
    stats = {'items_flagged': len(items_flagged_list),
             'items_checked': len(monitor_log_profiles) if monitor_log_profiles else 1}
    metadata = {"finding_name": "activity_log_retention_is_set_365_days_or_greater",
                "negative_name": "",
                "columns": ["Monitor Log Profile", "Retention Days"]}            
    return  {"items": items_flagged_list, "stats": stats, "metadata": metadata}

# 5.3, 5.4, 5.5, 5.6, 5.7, 5.8, 5.9, 5.10, 5.11, 5.12    
log_alert_policies_str = '''
- alert_name: 'create_policy_assignment'
  operation_name: 'Microsoft.Authorization/policyAssignments/write'
  present: False
- alert_name: 'create_or_update_network_security_group'
  operation_name: 'Microsoft.Network/networkSecurityGroups/write'
  present: False
- alert_name: 'delete_network_security_group'
  operation_name: 'Microsoft.Network/networkSecurityGroups/delete'
  present: False
- alert_name: 'create_or_update_network_security_group_rule'
  operation_name: 'Microsoft.Network/networkSecurityGroups/securityRules/write'
  present: False
- alert_name: 'delete_network_security_group_rule'
  operation_name: 'Microsoft.Network/networkSecurityGroups/securityRules/delete'
  present: False
- alert_name: 'create_or_update_security_solution'
  operation_name: 'Microsoft.Security/securitySolutions/write'
  present: False
- alert_name: 'delete_security_solution'
  operation_name: 'Microsoft.Security/securitySolutions/delete'
  present: False
- alert_name: 'update_or_create_SQL_server_firewall_rule'
  operation_name: 'Microsoft.Sql/servers/firewallRules/write'
  present: False
- alert_name: 'delete_SQL_server_firewall_rule'
  operation_name: 'Microsoft.Sql/servers/firewallRules/delete'
  present: False
- alert_name: 'update_security_policy'
  operation_name: 'Microsoft.Security/policies/write'
  present: False
'''
log_alert_policies = yaml.load(log_alert_policies_str)

def activity_log_alert_is_configured(activity_log_alerts, log_alert_policies):
    """
    #TODO WIP
    For each resource_group determine if activity-log alerts are configured correctly
    @returns: list of [resource_group, True of False for 5.3 to 5.12 in succession]
    """
    items_flagged_list = []

  
    for log_alert in activity_log_alerts:
        condition = log_alert.get('condition', [])
        if not condition:
            continue
        conditions = condition.get('allOf', [])
        if not conditions:
            continue
        for condition in conditions:
            for log_alert_policy in log_alert_policies:
                if condition.get('equals') and (condition.get('equals') == log_alert_policy["operation_name"]):
                    log_alert_policy["present"] = True

    for log_alert_policy in log_alert_policies:
        if log_alert_policy['present'] == False:
            items_flagged_list.append((log_alert_policy['alert_name'], log_alert_policy['operation_name']))
    
    stats = {'items_flagged': len(items_flagged_list),
             'items_checked': len(log_alert_policies)}
    metadata = {"finding_name": "activity_log_alert_is_configured",
                "negative_name": "",
                "columns": ["Missing Policy"]}            
    return  {"items": items_flagged_list, "stats": stats, "metadata": metadata}
                                                                    
#@gen_results(results)
MIN_ACTIVITY_LOG_RETENDION_DAYS = 365
MIN_KEY_VAULT_RETENTION_DAYS = 180
def logging_for_azure_keyvault_is_enabled_5_13(resource_diagnostic_settings):        
    items_flagged_list = []
    for setting in resource_diagnostic_settings:
        keyvault_settings_values = setting['value']
        if keyvault_settings_values:
            for value in keyvault_settings_values:
                # Do we need to loop over ['logs'] list as well?  My lists are length 1, only checking [0]
                keyvault_name = value['keyvault_name']
                resource_group = value['resourceGroup']
                if not value.get('logs', None):
                    enabled = False
                    retention_enabled = False
                    retention_days = 0
                else:
                    enabled = value['logs'][0]['enabled']
                    retention_enabled = value['logs'][0]['retentionPolicy']['enabled']
                    retention_days = value['logs'][0]['retentionPolicy']['days']                    
                if not (enabled and retention_enabled and (retention_days >= MIN_KEY_VAULT_RETENTION_DAYS)):
                    items_flagged_list.append((keyvault_name, enabled, retention_enabled, retention_days))
        else:
            items_flagged_list.append((resource_group, keyvault_name, "False", "False", "None"))
            
    stats = {'items_flagged': len(items_flagged_list),
             'items_checked': len(resource_diagnostic_settings)}
    metadata = {"finding_name": "logging_for_azure_keyvault_is_enabled",
                "negative_name": "",
                "columns": ["Keyvault", "Enabled", "Retention Enabled", "Retention Days"]}            
    return  {"items": items_flagged_list, "stats": stats, "metadata": metadata}
              
    



import datetime
import os
import subprocess
import sys
import re
import json
import requests
import traceback
import yaml

# azure_cis_scanner.credentials is a modification of https://github.com/Azure/azure-sdk-for-python/blob/master/azure-common/azure/common/credentials.py
# until https://github.com/Azure/azure-sdk-for-python/issues/2898 gets fixed
from azure.common.client_factory import get_client_from_cli_profile, get_client_from_auth_file
from azure.mgmt.compute import ComputeManagementClient
from azure.mgmt.sql import SqlManagementClient
from azure.mgmt.monitor import MonitorManagementClient
from azure.mgmt.network import NetworkManagementClient
from azure.mgmt.monitor.models import RuleMetricDataSource
from azure.mgmt.resource import ResourceManagementClient, SubscriptionClient
from azure.common.credentials import ServicePrincipalCredentials

from azure_cis_scanner.credentials import get_azure_cli_credentials

AZURE_CONFIG_DIR = os.path.expanduser('~/.azure')
AZURE_PROFILE_PATH = os.path.join(AZURE_CONFIG_DIR, 'azureProfile.json')
AZURE_CREDENTIALS_PATH = os.path.join(AZURE_CONFIG_DIR, 'credentials')
AZURE_SERVICE_PRINCIPALS_PATH = os.path.join(AZURE_CONFIG_DIR, 'servicePrincipals-{subscription_id}.json')

# https://github.com/Azure/azure-cli/blob/dev/src/command_modules/azure-cli-keyvault/azure/cli/command_modules/keyvault/_client_factory.py
def keyvault_data_plane_factory(cli_ctx, _):
    from azure.keyvault import KeyVaultAuthentication, KeyVaultClient
    from azure.cli.core.profiles import ResourceType, get_api_version
    version = str(get_api_version(cli_ctx, ResourceType.DATA_KEYVAULT))

    def get_token(server, resource, scope):  # pylint: disable=unused-argument
        import adal
        from azure.cli.core._profile import Profile
        try:
            return Profile(cli_ctx=cli_ctx).get_raw_token(resource)[0]
        except adal.AdalError as err:
            from knack.util import CLIError
            # pylint: disable=no-member
            if (hasattr(err, 'error_response') and
                    ('error_description' in err.error_response) and
                    ('AADSTS70008:' in err.error_response['error_description'])):
                raise CLIError(
                    "Credentials have expired due to inactivity. Please run 'az login'")
            raise CLIError(err)

    return KeyVaultClient(KeyVaultAuthentication(get_token), api_version=version)

def get_service_principal_credentials(subscription_id, auth_type='sdk', refresh_sp_credentials=False):
    """
    Get service principal credentials required by KeyVault, Storage 
    """
    sp_path = AZURE_SERVICE_PRINCIPALS_PATH.format(subscription_id=subscription_id)
    
    if auth_type == 'sdk':
        if refresh_sp_credentials:
            if os.path.exists(sp_path):
                os.remove(sp_path)
        if os.path.exists(sp_path):
            with open(sp_path, 'r') as f:
                credentials = json.load(f)
        else:
            credentials = json.loads(call("az ad sp create-for-rbac --sdk-auth", stderr=None))
        sp_credentials = ServicePrincipalCredentials(
            client_id=credentials['clientId'],
            secret=credentials['clientSecret'],
            tenant=credentials['tenantId']
        )
        with open(sp_path, 'w') as f:
            json.dump( dict(clientId=credentials['clientId'],
                            clientSecret=credentials['clientSecret'],
                            tenantId=credentials['tenantId']),
                         f, indent=4, sort_keys=True)
    
    return sp_credentials

def get_credentials_from_cli(tenant_id=None, subscription_id=None):
    """
    Create a credential for each of the subscriptions in azureProfile.json for a tenant
    @param tenant_id: uuid string - if None, iterate over all tenant_ids
    @param subscription_id: uuid string - if None iterate over all subscription_ids
    @returns: list of (tenant_id, subscription_id, subscription_name, credentials) where credentials is an ADAL signed session 
              bound to the tenant_id and subscription
    """

    with open(AZURE_PROFILE_PATH, 'r') as f:
        azure_profiles = yaml.load(f)['subscriptions']
    results = []
    for profile in azure_profiles:
        if tenant_id and not (tenant_id == profile['tenantId']):
            continue
        tenant_id = profile['tenantId']
        if subscription_id and (subscription_id != profile['id']):
            continue
        subscription_id = profile['id']
        subscription_name = profile['name']
        service_principle_name = profile['name'] + '-' + subscription_id
        
        # this is a modification of https://github.com/Azure/azure-sdk-for-python/blob/master/azure-common/azure/common/credentials.py
        # until https://github.com/Azure/azure-sdk-for-python/issues/2898 gets fixed
        credentials = get_azure_cli_credentials(resource=None, with_tenant=False, subscription_id=subscription_id)[0]
        
        results.append((tenant_id, subscription_id, subscription_name, credentials))
    return results

def get_clients_from_cli(subscription_id):
    credentials, subscription_id, tenant_id = get_azure_cli_credentials(resource=None, with_tenant=True, subscription_id=subscription_id)

    print("creating subscription client")
    subscription_client = SubscriptionClient(credentials)
    print("creating compute client")
    compute_client = ComputeManagementClient(credentials, subscription_id)
    print("creating sql client")
    sql_client = SqlManagementClient(credentials, subscription_client)
    print("creating resource manager client")
    rm_client = ResourceManagementClient(credentials, subscription_id)

    return subscription_client, compute_client, rm_client, sql_client

def get_list_from_paged_results(paged):
    results = []
    results.extend(paged.advance_page())
    while paged.next_link:
        results.extend(paged.advance_page)
    return [x.as_dict() for x in results]

def get_clients_from_service_principals(tenant_id=None, generate_credentials_ini=False, generate_auth_file=False, overwrite_ini=False):
    """
    Create a client for each of the subscriptions in azureProfile.json for a tenant
    """
    credentials_ini = ""

    with open(AZURE_PROFILE_PATH, 'r') as f:
        azure_profiles = yaml.load(f)['subscriptions']
    for profile in azure_profiles:
        if tenant_id and not (tenant_id == profile['tenantId']):
            continue
        subscription_id = profile['id']
        service_principle_name = profile['name'] + '-' + subscription_id
        result = call("az account set --subscription {}".format(subscription_id))
        credentials = jsonify(call("az ad sp create-for-rbac --sdk-auth"))

        credentials_ini += credentials_block(subscription_id, credentials, service_principle_name)
        if generate_auth_file:
            azure_auth_location = os.path.join(AZURE_CONFIG_DIR, service_principle_name + '.json')
            with open(azure_auth_location, 'w') as f:
                f.write(credentials)

        sp_credentials = ServicePrincipalCredentials(
            client_id=credentials['clientId'],
            secret=credentials['clientSecret'],
            tenant=credentials['tenantId']
        )
        if generate_credentials_ini:
            if overwrite_ini:
                mode = 'w'
            else:
                mode = 'w+'
            with open(AZURE_CREDENTIALS_PATH, mode) as f:
                f.write(credentials_ini)

        subscription_client = SubscriptionClient(credentials, subscription_id)
        compute_client = ComputeManagementClient(credentials, subscription_id)
        sql_client = SqlManagementClient(credentials, subscription_id)

        yield subscription_client, compute_client, sql_client


def credentials_block(subscription_id, credentials, service_principal_name, is_default=False):
    """
    Generate a block for use in ~/.azure/credentials ini file
    """
    client_id = credentials['client']
    secret = credentials['secret']
    tenant_id = credentials['tenant']

    ini_content = ""

    if is_default:
        ini_content += """ 
[default]
subscription_id={}
client_id={}
secret={}
tenant={}
""".format(subscription_id, client_id, secret, tenant_id)

    else:
        ini_content += """ 
[{}]
subscription_id={}
client_id={}
secret={}
tenant={}
""".format(service_principal_name, subscription_id, client_id, secret, tenant_id)
    return ini_content


def get_resource_groups(client, subscription_id, resource_groups_path):
    groups = []
    rgs = client.resource_groups.list()
    groups.extend(rgs.advance_page())
    while rgs.next_link:
        groups.extend(rgs.advance_page)
    resource_groups = [x.as_dict() for x in groups]

    with open(resource_groups_path, 'w') as f:
        json.dump(resource_groups, f, indent=4, sort_keys=True)
    return resource_groups

def load_resource_groups(resource_groups_path):
    with open(resource_groups_path, 'r') as f:
        resource_groups = yaml.safe_load(f)
    return resource_groups

token_expiry = None
access_token = None
filtered_data_dir = ''
scan_data_dir = ''
raw_data_dir = ''

def set_scans_dir(scans_dir):
    """
    Set scans_dir to requested value if it exists, otherwise default to ./scans and create
    """
    if not os.path.exists(scans_dir):
        if os.path.exists(os.path.expanduser('~/engagements/cis_test/scans')):
            scans_dir = '~/engagements/cis_test/scans'    
        elif os.path.exists('/engagements/cis_test/scans'):
            scans_dir = '/engagements/cis_test/scans'                
        else:
            scans_dir = os.path.join(os.getcwd(), 'scans')
            if not os.path.exists(scans_dir):
                os.mkdir(scans_dir)
        print("scans_dir {} not found.  Using {}".format(parser.scans_dir, scans_dir))
    return scans_dir

def set_data_paths(subscription_dirname, base_dir='.'):
    """
    Given a base_dir, create subdirs scans/{day}/raw
                                          /filtered
    @returns: scan_data_dir, raw_data_dir
    """
    # Get day in YYYY-MM-DD format

    day = datetime.datetime.now().strftime('%Y-%m-%d')

    if base_dir.startswith('/'):
        base_dir = os.path.abspath(base_dir)
    elif base_dir.startswith('~'):
        base_dir = os.path.expanduser(base_dir)
    scan_data_dir = os.path.join(base_dir, subscription_dirname, day)
    print("scan_data_dir", scan_data_dir)
    raw_data_dir = scan_data_dir + '/raw'
    print("raw_data_dir", raw_data_dir)
    if not os.path.exists(raw_data_dir):
        os.makedirs(raw_data_dir)
    filtered_data_dir = scan_data_dir + '/filtered'
    print("filtered_data_dir", filtered_data_dir)
    if not os.path.exists(filtered_data_dir):
        os.makedirs(filtered_data_dir)
    return scan_data_dir, raw_data_dir, filtered_data_dir   


def call(command, retrieving_access_token=False, stderr=None):
    #if not valid_token() and not retrieving_access_token:
    #    get_access_token()
    if(isinstance(command, str)) :
        command = command.split()           # subprocess needs an array of arguments
    try :
        print('running: ', command)
        result = subprocess.check_output(command, shell=False, stderr=stderr).decode('utf-8')
        #print("result", result)
        return result
    # allow calling code to raise AzScannerException and continue
    except AzScannerException as e:
        print("An exception occurred while processing command " + str(command) )
        print(e)
        print(traceback.format_exc())
        raise(e)


def verify_subscription_id_format(subscription_id) :
    r = re.compile("([a-f]|[0-9]){8}-([a-f]|[0-9]){4}-([a-f]|[0-9]){4}-([a-f]|[0-9]){4}-([a-f]|[0-9]){12}")
    if r.match(subscription_id):
        return True
    else :
        return False


def valid_token():
     if (not token_expiry) or (datetime.datetime.utcnow() > token_expiry):
        return False
     else:
        return True


def get_active_account():
    return jsonify(call("az account show"))


def get_subscription_id(account) :
    return account["id"]


def get_subscription_name(subscription_id, accounts):
    for account in accounts:
        if subscription_id == accounts['id']:
            return accounts['name']
    raise ValueError("subscription_id {} not found in accounts {}".format(subscription_id, accounts))

def get_subscription_dirname(subscription_id, subscription_name):
    return subscription_name.split(' ')[0] + '-' + subscription_id.split('-')[0]

def get_access_token():
    global token_expiry, access_token
    if not valid_token():
        complete_token = call("az account get-access-token", retrieving_access_token=True)
        complete_token = jsonify(complete_token)
        access_token = complete_token["accessToken"]
        token_expiry = complete_token["expiresOn"]
        print(token_expiry)
        token_expiry = datetime.datetime.strptime(token_expiry, '%Y-%m-%d %H:%M:%S.%f')
    return access_token, token_expiry

def set_credentials_tuples(parser):

    # While we have a mixture of azcle and az-python-sdk we need to set the subscription_id 
    # both with az account set, and with subscription_client using a modified credentials.py
    use_api_for_auth = parser.use_api_for_auth

    if parser.subscription_id:
        if verify_subscription_id_format(parser.subscription_id):
            subscription_id = parser.subscription_id
            try:
                if use_api_for_auth:
                    credentials_tuples = get_credentials_from_cli(subscription_id=subscription_id)
                    tenant_id, subscription_id, subscription_name, credentials = credentials_tuples[0]
                    if parser.tenant_id and (tenant_id != parser.tenant_id):
                        raise(ValueError("subscription {} does not belong to tenant {}".format(subscription_id, parser.tenant_id))) 

                else:
                    call("az account set --subscription {}".format(subscription_id))
                    account = json.loads(call("az account show"))
                    if parser.tenant_id:
                        if account['tenantId'] != parser.tenant_id:
                            raise(ValueError("subscription {} does not belong to tenant {}".format(subscription_id, parser.tenant_id))) 
            except Exception as e:
                print(e)
                print(traceback.format_exc())
                raise ValueError("Unable to set or find subscription {}".format(subscription_id))
        else:
            raise ValueError("supplied subscription id '{}' is invalid".format(parser.subscription_id))
    else:
        try:
            if parser.tenant_id:
                if not use_api_for_auth:
                    raise(ValueError("use_api_for_auth=False is only supported for default tenant.  Create an issue or PR"))
                tenant_id = parser.tenant_id
                print("No subscription specified, running on all subscriptions for supplied tenant {}".format(tenant_id))
                credentials_tuples = get_credentials_from_cli(tenant_id=tenant_id)
                print(credentials_tuples)
            else:
                print("No tenant or subscription specified, getting active account")
                account = get_active_account()
                print('account', account)
                subscription_id = account['id']
                credentials_tuples = get_credentials_from_cli(subscription_id=subscription_id)

        except Exception as e:
            print(e)
            print(traceback.format_exc())
            print("No Azure account associated with this session. Please authenticate to continue.")
            call("az login")
            account = get_active_account()

    if not use_api_for_auth:
        subscription_id = account['id']
        subscription_name = account['name']
        print("Using subscription_id {} {}".format(subscription_id, subscription_name))
        print("Re-run with --subscription-id if you wish to change")
        credentials_tuples = get_credentials_from_cli(subscription_id=subscription_id)

    return credentials_tuples

def make_request(url, args=[]):
    print('requesting ', url)
    authorization_headers = {"Authorization" : "Bearer " + get_access_token()[0]}
    r = requests.get(url, headers=authorization_headers)
    return r.text


def jsonify(jsonString) :
    return json.loads(jsonString)


def stringify(jsonObject) :
    return json.dumps(jsonObject)


class AzScannerException(Exception):
    def __init__(self,*args,**kwargs):
        Exception.__init__(self,*args,**kwargs)




import datetime
import json
import os
import traceback
import yaml

from azure.mgmt.network import NetworkManagementClient

from azure_cis_scanner import utils
from azure_cis_scanner.utils import get_list_from_paged_results, AzScannerException

resource_groups_path = os.path.join(config['raw_data_dir'], "resource_groups.json")
network_flows_path = os.path.join(config['raw_data_dir'], "network_flows.json")
network_watcher_path = os.path.join(config['raw_data_dir'], "network_watcher.json")
networking_filtered_path = os.path.join(config['filtered_data_dir'], 'networking_filtered.json')
network_security_groups_path = os.path.join(config['raw_data_dir'], "network_security_groups.json")
credentials = config['cli_credentials']
subscription_id = config['subscription_id']

##########################
# Get Raw Data
##########################

def get_data():
    network_security_groups = get_network_security_groups(network_security_groups_path)
    get_network_watcher(network_watcher_path)
    get_network_flows(network_flows_path, network_security_groups)
    
def get_network_security_groups(network_security_groups_path):
    """
    @network_path: string - path to output json file
    """
    network_security_groups = json.loads(utils.call("az network nsg list"))
    with open(network_security_groups_path, 'w') as f:
        json.dump(network_security_groups, f, indent=4, sort_keys=True)
    return network_security_groups

def load_network_security_groups(network_security_groups_path):
    with open(network_security_groups_path, 'r') as f:
        network_security_groups = yaml.load(f)
    return network_security_groups

approved_regions = []
def get_network_watcher(network_watcher_path):
    """
    @network_watcher_path: string - path to output json file
    """
    network_watcher = json.loads(utils.call("az network watcher list"))
    with open(network_watcher_path, 'w') as f:
        json.dump(network_watcher, f, indent=4, sort_keys=True)
    return network_watcher

def load_network_watcher(network_watcher_path):
    with open(network_watcher_path, 'r') as f:
        network_watcher = yaml.load(f)
    return network_watcher

def get_network_flows(network_flows_path, network_security_groups):
    """
    @network_flows_path: string - path to output json file
    @network_security_groups: list of nsgs
    @returns: list of network flow dicts
    """
    network_flows = []
    for nsg in network_security_groups:
        resource_group = nsg['resourceGroup']
        nsg_id = nsg['id']
        network_flow = json.loads(utils.call("az network watcher flow-log show --resource-group {resource_group} --nsg {nsg_id}".format(
            resource_group=resource_group, nsg_id=nsg_id)))
        nsg_name = nsg["name"]
        network_flows.append({"resource_group": resource_group, "nsg_name": nsg_name, "network_flow": network_flow})
        
    with open(network_flows_path, 'w') as f:
        json.dump(network_flows, f, indent=4, sort_keys=True)
    return network_flows

def load_network_flows(network_flows_path):
    with open(network_flows_path, 'r') as f:
        network_flows = yaml.load(f)
    return network_flows

##########################
# Tests
##########################

def test_controls():
    """
    Generate filtered (failing) output in json
    """
    network_watcher = load_network_watcher(network_watcher_path)
    network_security_groups = load_network_security_groups(network_security_groups_path)
    resource_groups = utils.load_resource_groups(resource_groups_path)
    network_flows = load_network_flows(network_flows_path)
    networking_results = {}

    networking_results['access_is_restricted_from_the_internet'] = access_is_restricted_from_the_internet_6_1(network_security_groups)
    networking_results['network_security_group_flow_log_retention_period_is_greater_than_90_days'] = network_security_group_flow_log_retention_period_is_greater_than_90_days_6_4(network_flows)
    networking_results['network_watcher_is_enabled'] = network_watcher_is_enabled_6_5(network_watcher)
                
    with open(networking_filtered_path, 'w') as f:
        json.dump(networking_results, f, indent=4, sort_keys=True)
    return networking_results

# 6.1, 6.2, 
def access_is_restricted_from_the_internet_6_1(network_security_groups):
    items_flagged_list = []
    for nsg in network_security_groups:
        # should actually be any port range that includes 3389
        security_rules = nsg['securityRules']
        for security_rule in security_rules:
            if security_rule['destinationPortRange'] == '3389' and security_rule['direction'] == 'Inbound' and security_rule['protocol'] == 'TCP':
                if security_rule['sourceAddressPrefix'] in ['*', '/0', 'internet', 'any']:
                    items_flagged_list.append((nsg['resourceGroup'],nsg['name'], '3389', security_rule))
            if security_rule['destinationPortRange'] == '22' and security_rule['direction'] == 'Inbound':
                if security_rule['sourceAddressPrefix'] in ['*', '/0', 'internet', 'any']:
                    items_flagged_list.append((nsg['resourceGroup'],nsg['name'], '22', security_rule['name']))
                    
    stats = {'items_flagged': len(items_flagged_list),
             'items_checked': len(network_security_groups)}
    metadata = {"finding_name": "access_is_restricted_from_the_internet",
                "negative_name": "",
                "columns": ["Resource Group", "NSG", "Port", "Rule"]}            
    return  {"items": items_flagged_list, "stats": stats, "metadata": metadata}


def sql_server_access_is_restricted_from_the_internet_6_3():
    """
    Powershell
    """
    pass                

def network_security_group_flow_log_retention_period_is_greater_than_90_days_6_4(network_flows):
    items_flagged_list = []
    for network_flow in network_flows:
        flow = network_flow['network_flow']
        if flow['enabled'] == False:
            status = "Not enabled"
            items_flagged_list.append((network_flow['resource_group'], network_flow['nsg_name'], status))
        elif flow['retentionPolicy']['days'] == 0:
            continue
        elif (flow['retentionPolicy']['days'] < 90) or (flow['retentionPolicy']['enabled'] == False):
            status("Days {}, Enabled {}".format(flow['retentionPolicy']['days'], flow['retentionPolicy']['enabled']))
            items_flagged_list.append((network_flow['resource_group'], network_flow['nsg_name'], status))

    stats = {'items_flagged': len(items_flagged_list),
             'items_checked': len(network_flows)}
    metadata = {"finding_name": "network_security_group_flow_log_retention_period_is_greater_than_90_days",
                "negative_name": "",
                "columns": ["Resource Group", "Network Flow", "Status"]}            
    return  {"items": items_flagged_list, "stats": stats, "metadata": metadata}

def network_watcher_is_enabled_6_5(network_watcher):
    items_flagged_list = []    
    for watcher in network_watcher:
        if watcher['provisioningState'] != 'Succeeded':
            items_flagged_list.append((watcher))
            
    stats = {'items_flagged': len(items_flagged_list),
             'items_checked': len(network_watcher)}
    metadata = {"finding_name": "network_security_group_flow_log_retention_period_is_greater_than_90_days",
                "negative_name": "",
                "columns": ["Resource Group", "Network Flow", "Status"]}            
    return  {"items": items_flagged_list, "stats": stats, "metadata": metadata}



import json
import os
import yaml

from azure.mgmt.sql import SqlManagementClient
from azure_cis_scanner import utils

credentials = config['cli_credentials']
sp_credentials = config['sp_credentials']
subscription_id=config['subscription_id']

filtered_sql_servers_path = os.path.join(config['filtered_data_dir'], 'sql_servers_filtered.json')
sql_servers_path = os.path.join(config['raw_data_dir'], 'sql_databases.json')
sql_databases_path = os.path.join(config['raw_data_dir'], 'sql_servers.json')
sql_server_policies_path = os.path.join(config['raw_data_dir'], 'sql_server_policies.json')

import os
import yaml

def get_data():
    sql_servers = get_sql_servers(sql_servers_path)
    sql_databases = get_sql_databases(sql_databases_path, sql_servers)
    get_sql_server_policies(sql_server_policies_path, sql_servers)

def get_sql_servers(sql_servers_path) :
    sql_servers = json.loads(utils.call("az sql server list"))
    with open(sql_servers_path, 'w') as f:
        json.dump(sql_servers, f, indent=4, sort_keys=True)
    return sql_servers

def get_sql_databases(sql_databases_path, sql_servers) :
    sql_dbs = []
    for sql_server in sql_servers:
        server_name = sql_server['name']
        resource_group = sql_server['resourceGroup']
        dbs = json.loads(utils.call("az sql db list --resource-group {} --server {}".format(resource_group, server_name)))
        sql_dbs.extend(dbs)
    with open(sql_databases_path, 'w') as f:
        json.dump(sql_dbs, f, indent=4, sort_keys=True)
    return sql_dbs

def get_sql_server_policies(sql_server_policies_path, sql_servers):
    results = {}
    for sql_server in sql_servers:
        server_name = sql_server['name']
        resource_group = sql_server['resourceGroup']
        sql_server_policies = {}
        sql_server_policies['audit_policy'] = get_sql_server_audit_policies(subscription_id, resource_group, server_name)
        sql_server_policies['threat_detection_policy'] = get_sql_server_threat_detection_policies(subscription_id, resource_group, server_name)
        sql_server_policies['active_directory_admin_configurations'] = get_sql_server_active_directory_admin_configuration(subscription_id, resource_group, server_name)
        results[(resource_group, server_name)] = sql_server_policies
    with open(sql_server_policies_path, 'w') as f:
        yaml.dump(results, f)
    return results

def load_sql_servers(sql_servers_path):
    with open(sql_servers_path, 'r') as f:
        sql_servers = yaml.load(f)
    return sql_servers

def load_sql_server_policies(sql_server_policies_path):
    with open(sql_server_policies_path, 'r') as f:
        sql_server_policies = yaml.load(f)
    return sql_server_policies
    
# This function will be recentered around Azure Command Line, after such an option becomes available.
def get_sql_server_audit_policies(subscription_id, resource_group, server_name):
    endpoint = "https://management.azure.com/subscriptions/"+subscription_id+"/resourceGroups/"+resource_group+"/providers/Microsoft.Sql/servers/"+server_name+"/auditingSettings/Default?api-version=2015-05-01-preview"
    sql_server_audit_policy = utils.make_request(endpoint)
    sql_server_audit_policy = utils.jsonify(sql_server_audit_policy)
    return sql_server_audit_policy

# This function will be recentered around Azure Command Line, after such an option becomes available.
def get_sql_server_threat_detection_policies(subscription_id, resource_group, server_name):
    endpoint = "https://management.azure.com/subscriptions/"+subscription_id+"/resourceGroups/"+resource_group+"/providers/Microsoft.Sql/servers/"+server_name+"/securityAlertPolicies/Default?api-version=2015-05-01-preview"
    sql_server_threat_detection_policy = utils.make_request(endpoint)
    sql_server_threat_detection_policy = utils.jsonify(sql_server_threat_detection_policy)
    return sql_server_threat_detection_policy

def get_sql_server_active_directory_admin_configuration(subscription_id, resource_group, server_name):
    active_directory_admin_configuration = utils.call("az sql server ad-admin list --resource-group " + resource_group + " --server " + server_name)
    active_directory_admin_configuration = utils.jsonify(active_directory_admin_configuration)
    return active_directory_admin_configuration

##################
# Tests
##################
def wrap(pre, post):
    def decorate(func):
        def call(*args, **kwargs):
            pre(func, *args, **kwargs)
            result = func(*args, **kwargs)
            post(func, result, results, *args, **kwargs)
            return result
        return call
    return decorate

def remove_section_digits(name):
    filtered = []
    name_words = name.split('_')
    # remove trailing digits, taking care not to remove 90 in ...than_90_days_4_1_7
    for i, word in enumerate(name_words):
        if not word.isdigit() or ( (i < len(name_words)-2) and not name_words[i+1].isdigit()):
            filtered.append(word)
    return '_'.join(filtered)

def trace_in(func, *args, **kwargs):
    pass

def trace_out(func, result, *args, **kwargs):
    name = remove_section_digits(func.__name__)
    finding_results = results.get(name, {})
    if finding_results:
        items_flagged_list = finding_results["items"]
        items_checked = finding_results["stats"]["items_checked"]
    else:
        items_flagged_list = []
        items_checked = 0
    items_checked += 1
    if not result:
        items_flagged_list.append((kwargs['resource_group'], kwargs['server_name']))
           
    results[name] = {"items": items_flagged_list, "stats": {"items_checked": items_checked}}
        
results = {}

def test_controls() :
    global results
    sql_servers = load_sql_servers(sql_servers_path)
    sql_server_policies = load_sql_server_policies(sql_server_policies_path)
    
    for (resource_group, server_name), sql_server_policy in sql_server_policies.items():
        sql_server_audit_policy = sql_server_policies[(resource_group, server_name)]['audit_policy']
        sql_server_threat_detection_policy = sql_server_policies[(resource_group, server_name)]['threat_detection_policy']
        sql_server_active_directory_admin_configurations = sql_server_policies[(resource_group, server_name)]['active_directory_admin_configurations']

        auditing_is_set_to_on_4_1_1(sql_server_audit_policy, resource_group=resource_group, server_name=server_name)
        threat_detection_is_set_to_on_4_1_2(sql_server_threat_detection_policy, resource_group=resource_group, server_name=server_name)
        threat_detection_types_is_set_to_all_4_1_3(sql_server_threat_detection_policy, resource_group=resource_group, server_name=server_name)
        send_alerts_to_is_set_4_1_4(sql_server_threat_detection_policy, resource_group=resource_group, server_name=server_name)
        email_service_and_co_administrators_is_enabled_4_1_5(sql_server_threat_detection_policy, resource_group=resource_group, server_name=server_name)
        auditing_retention_is_greater_than_90_days_4_1_6(sql_server_audit_policy, resource_group=resource_group, server_name=server_name)
        threat_detection_retention_is_greater_than_90_days_4_1_7(sql_server_threat_detection_policy, resource_group=resource_group, server_name=server_name)
        azure_active_directory_admin_is_configured_4_1_8(sql_server_active_directory_admin_configurations, resource_group=resource_group, server_name=server_name)

    stats_results = {}
    for finding in results:
        items_flagged_list = results[finding]["items"]
        items_checked = results[finding]["stats"]["items_checked"]
        items_flagged = len(items_flagged_list)
        stats = {'items_flagged': len(items_flagged_list),
                 'items_checked': items_checked}
        metadata = {"finding_name": finding,
                    "negative_name": "",
                    "columns": ["Region", "Server"]}            
        stats_results[finding] = {"items": items_flagged_list, "stats": stats, "metadata": metadata}
        
    with open(filtered_sql_servers_path, 'w') as f:
        yaml.dump(stats_results, f)
    # clear results for next run
    results = {}
    return stats_results

@wrap(trace_in, trace_out)
def auditing_is_set_to_on_4_1_1(sql_server_audit_policies, resource_group=None, server_name=None):
    if sql_server_audit_policies["properties"]["state"] == "Disabled" :
        return False
    else:
        return True

@wrap(trace_in, trace_out)
def threat_detection_is_set_to_on_4_1_2(sql_server_threat_detection_policies, resource_group=None, server_name=None):
    if sql_server_threat_detection_policies["properties"]["state"] == "Disabled" :
        return False
    else:
        return True
    
@wrap(trace_in, trace_out)
def threat_detection_types_is_set_to_all_4_1_3(sql_server_threat_detection_policies, resource_group=None, server_name=None):
    if sql_server_threat_detection_policies["properties"]["state"] == "Disabled" or sql_server_threat_detection_policies["properties"]["disabledAlerts"] != "":
        return False
    else:
        return True

@wrap(trace_in, trace_out)
def send_alerts_to_is_set_4_1_4(sql_server_threat_detection_policies, resource_group=None, server_name=None):
    if sql_server_threat_detection_policies["properties"]["state"] == "Disabled" or sql_server_threat_detection_policies["properties"]["emailAddresses"] != "":
        return False
    else:
        return True

@wrap(trace_in, trace_out)
def email_service_and_co_administrators_is_enabled_4_1_5(sql_server_threat_detection_policies, resource_group=None, server_name=None):
    if sql_server_threat_detection_policies["properties"]["state"] == "Disabled" or sql_server_threat_detection_policies["properties"]["emailAccountAdmins"] != "":
        return False
    else:
        return True

@wrap(trace_in, trace_out)
def auditing_retention_is_greater_than_90_days_4_1_6(sql_server_audit_policies, resource_group=None, server_name=None):
    if (sql_server_audit_policies["properties"]["state"] == "Disabled"):
        return False
    retention_days = int(sql_server_audit_policies["properties"]["retentionDays"])
    if (retention_days) ==0 or (retention_days > 90):
        return True
    else:    
        return False
    
@wrap(trace_in, trace_out)
def threat_detection_retention_is_greater_than_90_days_4_1_7(sql_server_threat_detection_policies, resource_group=None, server_name=None):
    if (sql_server_threat_detection_policies["properties"]["state"] == "Disabled"):
        return False
    retention_days = int(sql_server_threat_detection_policies["properties"]["retentionDays"])
    if (retention_days) ==0 or (retention_days > 90):
        return True
    else:    
        return False
    
@wrap(trace_in, trace_out)
def azure_active_directory_admin_is_configured_4_1_8(sql_server_active_directory_admin_configurations, resource_group=None, server_name=None):
    if not sql_server_active_directory_admin_configurations:
        return False
    else:
        return True


import datetime
import json
import os
import traceback
import yaml

from azure.mgmt.keyvault import KeyVaultManagementClient
from azure.keyvault import KeyVaultClient, KeyVaultAuthentication
from azure_cis_scanner import utils
from msrestazure.azure_active_directory import ServicePrincipalCredentials
#from azure.common.credentials import ServicePrincipalCredentials

from azure_cis_scanner.utils import get_list_from_paged_results, get_service_principal_credentials, AzScannerException
from azure.keyvault.models import KeyVaultErrorException

# RESOURCES
# https://github.com/Azure-Samples/key-vault-recovery-python/blob/master/key_vault_sample_config.py

keyvaults_path = os.path.join(config['raw_data_dir'], 'keyvaults.json')
keyvault_keys_and_secrets_metadata_path = os.path.join(config['raw_data_dir'], 'keyvault_keys_and_secrets_metadata.json')
locked_resources_path = os.path.join(config['raw_data_dir'], 'locked_resources.json')
other_security_considerations_filtered_path = os.path.join(config['filtered_data_dir'], 'other_security_considerations_filtered.json')
credentials = config['cli_credentials']
sp_credentials = config['sp_credentials']
subscription_id = config['subscription_id']

def get_keyvault_client():
    return KeyVaultClient(sp_credentials)

#############
# DATA
#############

def get_keyvaults(keyvaults_path):
    """
    @keyvaults_path: string - path to output json file
    @returns: list of virtual_machines dict
    """
    kvm_client = KeyVaultManagementClient(credentials, subscription_id)
    keyvaults = get_list_from_paged_results(kvm_client.vaults.list())
    with open(keyvaults_path, 'w') as f:
        json.dump(keyvaults, f, indent=4, sort_keys=True)
    return keyvaults

def load_keyvaults(keyvaults_path):
    with open(keyvaults_path, 'r') as f:
        keyvaults = yaml.load(f)
    return keyvaults

def get_locked_resources():
    lock_list = json.loads(utils.call("az lock list"))

    with open(locked_resources_path, 'w') as f:
        json.dump(lock_list, f, indent=4, sort_keys=True)
    return lock_list

def load_locked_resources(locked_resources_path):
    with open(locked_resources_path, 'r') as f:
        locked_list = yaml.load(f)
    return locked_list

def get_keyvault_keys_and_secrets_metadata(keyvault_keys_and_secrets_metadata_path, keyvaults):
    metadata = {}
    kvm_client = KeyVaultManagementClient(credentials, subscription_id)
    kv_client = get_keyvault_client()
    for keyvault in keyvaults:        
        subsciption_id, resource_group, vault_name = parse_vault_id(keyvault['id'])
        vault_url = 'https://{vault_name}.vault.azure.net'.format(vault_name=vault_name)
        print('vault_url', vault_url)
        metadata[vault_name] = {}
        try:
            keys = json.loads(utils.call("az keyvault key list --vault-name {}".format(vault_name)))
            metadata[vault_name]['keys'] = keys
            #keys = kv_client.get_keys(vault_url)
            #metadata[vault_name]['keys'] = get_list_from_paged_results(keys)
        except KeyVaultErrorException as e:
            if str(e.response) == '<Response [401]>':
                metadata[vault_name]['keys'] = "ERROR UNAUTHORIZED"
            else:
                print(e.with_traceback)
                print(e.response)
                print(kv_client.config.credentials)         
                raise(Exception("Unexpected KeyVaultErrorException response for vault {}".format(vault_url)))

        try:
            secrets = json.loads(utils.call("az keyvault secret list --vault-name {}".format(vault_name)))
            metadata[vault_name]['secrets'] = secrets
            #secrets = kv_client.get_secrets(vault_url)    
            #metadata[vault_name]['secrets'] = get_list_from_paged_results(secrets)
        except KeyVaultErrorException as e:
            if str(e.response) == '<Response [401]>':
                metadata[vault_name]['secrets'] = "UNAUTHORIZED"
            else:
                print(e.with_traceback)
                print(e.response)
                raise(Exception("Unexpected KeyVaultErrorException response"))
        except AzScannerException as e:
            print("Error calling command {}".format(e.message))
            print(e.with_traceback)

    with open(keyvault_keys_and_secrets_metadata_path, 'w') as f:
        json.dump(metadata, f, indent=4, sort_keys=True)
    return metadata

def load_keyvault_keys_and_secrets_metadata(keyvault_keys_and_secrets_metadata_path):
    with open(keyvault_keys_and_secrets_metadata_path, 'r') as f:
        metadata = yaml.load(f)
    return metadata

def get_data():
    keyvaults = get_keyvaults(keyvaults_path)
    get_keyvault_keys_and_secrets_metadata(keyvault_keys_and_secrets_metadata_path, keyvaults)
    get_locked_resources()

def parse_vault_id(vault_id):
    """
    parse the returned vaults
    vaults.list() returns {'url': '/subscriptions/{subscriptionId}/resourceGroups/{resourceGroupName}/providers/Microsoft.KeyVault/vaults/{vaultName}'}
    """
    _, _, subscription_id, _, resource_group, *unimportant, vault_name  = vault_id.split('/')
    return subscription_id, resource_group, vault_name

###############
# TESTS
###############

MAX_EXPIRY_ROTATION_DAYS = 730
# 8.1 and 8.2
def expiry_date_is_set_on_all_keys_and_secrets(keyvault_keys_and_secrets_metadata):
    items_flagged_list = []
    items_checked = 0
    today = datetime.datetime.today()
    
    def get_key_or_secret_status(info):
        enabled = info['attributes']['enabled']
        created = datetime.datetime.strptime(info['attributes']['created'].split('T')[0], '%Y-%m-%d')
        expires = info['attributes']['expires']
        status = "ok"
        if expires:
            expires = datetime.datetime.strptime(expires.split('T')[0], '%Y-%m-%d')
            expiry_delta = expires - created
            if today > expires:
                satus = "expired"
            elif expiry_delta > datetime.timedelta(days=MAX_EXPIRY_ROTATION_DAYS):
                status = "exceeds max expiry days"
            # convert times back to a string for display
            expires = expires.strftime('%Y-%m-%d')
        else:
            status = "no expiry"
            expires = "None"
        created = created.strftime('%Y-%m-%d')            
            
        return status, created, expires
    
    for keyvault_name, metadata in keyvault_keys_and_secrets_metadata.items():
        for key_info in metadata['keys']:
            items_checked += 1
            if "ERROR" in key_info:
                print("ERROR", metadata['keys']["ERROR"])
                items_flagged_list.append((keyvault_name, "ACCESS_DENIED", "key", "N/A", "N/A", "N/A"))
                continue
            key_name = key_info['kid'].split('/')[-1]
            status, created, expires = get_key_or_secret_status(key_info)
            if status != "ok":
                items_flagged_list.append((keyvault_name, key_name, "key", status, created, expires))
                
        for secret_info in metadata['secrets']:
            items_checked += 1
            if "ERROR" in secret_info:
                print("ERROR",  metadata['secrets']["ERROR"])
                items_flagged_list.append((keyvault_name, "ACCESS_DENIED", "secret", "N/A", "N/A", "N/A"))
                continue
            secret_name = secret_info['id'].split('/')[-1]
            status, created, expires = get_key_or_secret_status(secret_info)                
            if status != "ok":
                items_flagged_list.append((keyvault_name, secret_name, "secret", status, created, expires))
    
    stats = {'items_flagged': len(items_flagged_list),
             'items_checked': items_checked}
    metadata = {"finding_name": "expiry_date_is_set_on_all_keys_and_secrets",
                "negative_name": "",
                "columns": ["KeyVault Name", "Name", "Type", "Status", "Created", "Expires"]}            
    return  {"items": items_flagged_list, "stats": stats, "metadata": metadata}

critical_resources_list = []
def resource_locks_are_set_for_mission_critical_azure_resources_8_3(critical_resources_list, locked_resources):
    """
    This finding needs some work.
    It is not clear from `az lock list` what resource is being locked.
    For now, best ignore comparison and flag an error if len(critical_resources_list) < len(locked_resources)
    """
    items_flagged_list = []
    critical_resources = set(critical_resources_list)
    if len(locked_resources) == 0 and len(critical_resources_list) == 0:
        stats = {'items_flagged': 1,
                 'items_checked': 1}
    else:
        stats = {'items_flagged': len(critical_resources_list) - len(locked_resources),
                 'items_checked': len(critical_resources_list)}
        # This is really the inverse of the usual items flagged list
        #items_flagged_list = list(critical_resources.intersection(set(locked_resources)))
        items_flagged_list = [(x['name'], x['id'], x['notes']) for x in locked_resources]
    metadata = {"finding_name": "resource_locks_are_set_for_mission_critical_Azure_resources",
                "negative_name": "",
                "columns": ["Lock Name", "Lock ID", "Notes"]}            
    return  {"items": items_flagged_list, "stats": stats, "metadata": metadata}

def test_controls():
    keyvault_keys_and_secrets_metadata = load_keyvault_keys_and_secrets_metadata(keyvault_keys_and_secrets_metadata_path)
    locked_resources = load_locked_resources(locked_resources_path)
    results = {}
    results['expiry_date_is_set_on_all_keys_and_secrets'] = expiry_date_is_set_on_all_keys_and_secrets(keyvault_keys_and_secrets_metadata)
    results['resource_locks_are_set_for_mission_critical_azure_resources'] = resource_locks_are_set_for_mission_critical_azure_resources_8_3(critical_resources_list, locked_resources)
    
    with open(other_security_considerations_filtered_path, 'w') as f:
        json.dump(results, f, indent=4, sort_keys=True)
    return results


import json
import os
import yaml
    
def clean_latex(tuple_entry):
    """
    Filter/escape problematic characters
    Our minerva pdf generator chokes on '_', '*', ...
    and possibly other things.
    """
    def _clean_latex(tuple_entry_string):
        processed = False
        for symbol in ['_', '*']:
            if symbol in tuple_entry_string:
                tuple_entry_string = tuple_entry_string.replace(symbol, '\\' + symbol)
                processed = True
        if processed:
            return '\\texttt{' + tuple_entry_string + '}'
        else:
            return tuple_entry_string

    return _clean_latex(str(tuple_entry))
    
    
def render_latex(resource_tuples, header, title):
    header = list(map(clean_latex, header))
    title = clean_latex(title)
    
    render_table_start(header, title)
    num_columns = len(header)
    if num_columns > 1:
        line = ' & '.join(['{}']*num_columns)
    else:
        line = '{}'
    line = line + ' \\\\'
    for resource_tuple in resource_tuples:
        print('    ' + line.format(*map(clean_latex, resource_tuple)))
    render_table_end(header)
    
def render_table_start(header, title):
    """
    Render latex table suitable for minerva rendering
    
            
    If the elements in the table are very long you can correct spacing with invisible text like this:
    {\color[HTML]{FFFFFF} {}} & {\color[HTML]{333333}{spacing}}{\color[HTML]{FFFFFF} {}}{\color[HTML]{333333}{spacing_2}} & {\color[HTML]{333333}{spacing_3}}{\color[HTML]{FFFFFF}{}} \\

    TODO: Paginate pages to break nicely over many pages.
    
    Some exceptions used to manually fix for specific cases:
    
    Text wrap some long arrays based on answer by zyy on
    https://tex.stackexchange.com/questions/54069/table-with-text-wrapping
    
    GROUPS column had variable length and this worked.
    \begin{tabular}{|l|>{\centering\arraybackslash}m{10cm}|}
    \hline
    \multicolumn{2}{|c|}{IAM Users without MFA} \\
    \rowcolor[HTML]{333333}
    {\color[HTML]{FFFFFF}USER NAME} & {\color[HTML]{FFFFFF}GROUPS} \\
    """
    num_columns = len(header)
    entries = ['\color[HTML]{FFFFFF}' + '{}'.format(clean_latex(x)) for x in header]
    if num_columns > 1: 
        line = '} & {'.join(entries)
    else:
        line = entries[0]
    columns_format = '{|' + '|'.join(['l']*num_columns) + '|}'
    print('\\begin{tabular}' + '{}'.format(columns_format) + '\n'
         '    ' + '\\hline\n' +
         '    ' + '\\multicolumn{' + str(num_columns) + '}' +
        '{|c|}' + '{' + title + '}' +  ' \\\\\n' +
        '    ' + '\\rowcolor[HTML]{333333}\n' +
        '    ' + '{' + line + '}' + ' \\\\' 
    )
    

def render_table_end(header):
    """
    Render table end
    """
    print('    \\hline\n' +
          '\\end{tabular}\n')


from azure_cis_scanner import utils
from azure_cis_scanner.utils import get_resource_groups, get_credentials_from_cli
import argparse
import logging
import json
import os
import subprocess
import traceback
import time
from importlib.util import spec_from_file_location, module_from_spec
from os.path import splitext, basename


from azure.common.client_factory import get_client_from_cli_profile, get_client_from_auth_file
from azure.mgmt.resource import ResourceManagementClient, SubscriptionClient

logger = logging.getLogger(__name__)
handler = logging.StreamHandler()
formatter = logging.Formatter(
        '%(asctime)s %(name)-12s %(levelname)-8s %(message)s')
handler.setFormatter(formatter)
logger.addHandler(handler)


def load_module(module_name, **kwargs):
    name = splitext(basename(module_name))[0]
    spec = spec_from_file_location(name, module_name)
    module = module_from_spec(spec)
    module.config = kwargs
    spec.loader.exec_module(module)
    return module

def dprint(*thing):
    print(str(thing))
    print(thing)


def ddprint(*thing):
    dprint(thing)
    print(type(thing))
    print(dir(thing))

MODULES_PATH = os.path.join(os.path.dirname(os.path.realpath(__file__)), 'modules')

def _get_modules(modules=None, skip_modules=None):
    """
    Returns specified modules, all modules or all modules - skipped modules
    """
    if modules and skip_modules:
        print("WARNING: both modules and skip_modules specified, ignoring skip_modules")
    all_modules = [x for x in os.listdir(MODULES_PATH) if x.endswith('.py')]
    if modules:
        intersect_modules = set(all_modules).intersection(set(modules))
        diff_modules = set(modules).difference(intersect_modules)
        if diff_modules:
            print("Warning, requested modules {} were not found.  Proceeding with {}".format(diff_modules, intersect_modules))
        return list(intersect_modules)
    if skip_modules:
        return list(set(all_modules).difference(set(skip_modules)))
    return all_modules
    
def _get_data(config):
    raw_data_dir, filtered_data_dir, cli_credentials, subscription_id = config['raw_data_dir'], \
        config['filtered_data_dir'], config['cli_credentials'], config['subscription_id']

    modules = _get_modules(config['modules'], config['skip_modules'])
    for module in modules:
        print('Getting data for {}'.format(module.strip('.py')))
        module_path = os.path.join(os.path.dirname(os.path.realpath(__file__)), 'modules', module)
        try:
            # mod = load_module(module_path, **dict(raw_data_dir=raw_data_dir, 
            #                                       filtered_data_dir=filtered_data_dir, 
            #                                       cli_credentials=cli_credentials,
            #                                       subscription_id=subscription_id))
            mod = load_module(module_path, **config)

            mod.get_data()
        except Exception as e:
            print("Exception was thrown! Unable to run get_data() for {}".format(module))
            print("Module path {}".format(module_path))
            print(e)
            print(traceback.format_exc())

def _test_controls(config):
    raw_data_dir, filtered_data_dir = config['raw_data_dir'], config['filtered_data_dir']
    modules = _get_modules(config['modules'], config['skip_modules'])
    for module in modules:
        print('Testing controls for {}'.format(module.strip('.py')))
        module_path = os.path.join(os.path.dirname(os.path.realpath(__file__)), 'modules', module)
        try:
            mod = load_module(module_path, **config)
            #mod = load_module(module_path, **dict(raw_data_dir=raw_data_dir, filtered_data_dir=filtered_data_dir))
            mod.test_controls()
        except Exception as e:
            print("Exception was thrown! Unable to run test_controls() for {}".format(module))
            print("Module path {}".format(module_path))
            print(e)
            print(traceback.format_exc())


def main():
    mainparser = argparse.ArgumentParser()
    mainparser.add_argument('--tenant-id', default=None, help='azure tenant id, if None, use default.  Scanner assumes different runs/project dirs for distinct tenants')
    mainparser.add_argument('--subscription-id', default=None, help='azure subscription id, if None, use default, if "all" use all subscriptions with default tenant')
    # TODO, set default in __init__.py or somewhere and make it windows compatible
    mainparser.add_argument('--scans-dir', default='/engagements/cis_test/scans', help='existing base dir of where to place or load files')
    mainparser.add_argument('--stages', default='data,test,report', help='comma separated list of steps to run in data,test')
    mainparser.add_argument('--modules', default=None, help='comma separated list of module names e.g. security_center.py')
    mainparser.add_argument('--skip-modules', default=[], help='comma separated list of module names to skip')
    mainparser.add_argument('--use-api-for-auth', default=True, help='if false, use azure cli calling subprocess, else use python-azure-sdk')
    mainparser.add_argument('--refresh-sp-credentials', action='store_true', help='refresh service principal creds needed for keyvault')
    mainparser.add_argument('--loglevel', default='info', help='loglevel in ["debug", "info", "warning", "error"]')
    parser = mainparser.parse_args()

    loglevel = parser.loglevel

    if loglevel != 'info':
        print("Checking to see if there is an Azure account associated with this session.")
        print("Running with arguments {}".format(parser))

    if loglevel == "debug":
        logger.setLevel(logging.DEBUG)
    elif loglevel == "info":
        logger.setlevel(logging.INFO)
    elif loglevel == "warning":
        logger.setlevel(logging.INFO)
    elif loglevel == "error":
        logger.setlevel(logging.ERROR)


    credentials_tuples = utils.set_credentials_tuples(parser)

    scans_dir = os.path.realpath(parser.scans_dir)
    scans_dir = utils.set_scans_dir(scans_dir)

        
    accounts_list = json.loads(utils.call("az account list"))
    with open(os.path.join(scans_dir, 'accounts.json'), 'w') as f:
        json.dump(accounts_list, f, indent=4, sort_keys=True)

    for tenant_id, subscription_id, subscription_name, credentials in credentials_tuples:

        sp_credentials = utils.get_service_principal_credentials(subscription_id, auth_type='sdk', refresh_sp_credentials=parser.refresh_sp_credentials)
        print("sp_credentials", sp_credentials, type(sp_credentials))
        logger.debug("DEBUGGER WORKS! running stages for {} {} {}".format(tenant_id, subscription_id, subscription_name))
        print("running stages for {} {} {}".format(tenant_id, subscription_id, subscription_name))
        
        access_token, token_expiry = utils.get_access_token()
        # create a part-friendly/part-uniquie-id name
        subscription_dirname = utils.get_subscription_dirname(subscription_id, subscription_name)

        scan_data_dir, raw_data_dir, filtered_data_dir = utils.set_data_paths(subscription_dirname, scans_dir)

        modules = parser.modules
        if modules:
            modules = modules.split(',')
        skip_modules = parser.skip_modules
        if skip_modules:
            skip_modules = skip_modules.split(',')

        config = dict(raw_data_dir=raw_data_dir, 
            filtered_data_dir=filtered_data_dir, 
            modules=modules,
            skip_modules=skip_modules,
            cli_credentials=credentials,
            sp_credentials=sp_credentials,
            subscription_id=subscription_id
            )

        ### Get some data common to most modules
        resource_groups_path = os.path.join(raw_data_dir, "resource_groups.json")
        rm_client = ResourceManagementClient(credentials, subscription_id)
        resource_groups = get_resource_groups(rm_client, subscription_id, resource_groups_path)

        stages = parser.stages.split(',')

        if 'data' in stages:
            _get_data(config)
        if 'test' in stages:
            print('test', config)
            _test_controls(config)


def az_login():
    from azure.common.client_factory import get_client_from_cli_profile
    from azure.mgmt.compute import ComputeManagementClient
    from azure.mgmt.resource import ResourceManagementClient, SubscriptionClient
    from msrestazure.azure_active_directory import MSIAuthentication

    client = get_client_from_cli_profile(SubscriptionClient)
    accounts = [ x.as_dict() for x in list(client.subscriptions.list())]
    print(accounts)


if __name__ == "__main__":
    main()




# TODO: Turn these into real tests


def test_incompatible_subscription_and_tenant():
    """
    azscanner(subscription_id, tenant_id) where subscription_id does not belong to tenant_id should fail with useful error
    """
    pass

def test_invalid_subscription_id():
    pass

def test_no_subscription_id():
    pass

def test_empty_azure_config_dir():
    """
    What happens if ~/.azure is empty
    """
    pass

def test_expired_credentials():
    """
    Correct behavior if session tokens have expired
    """
    pass

def test_specify_non_default_tenant():
    """
    If we have multiple tenants and pass the non-default, ensure we target the correct one.
    
    mock the azureProfiles.json and test that we get get_credentials_from_cli is called on the right data.
    create the parser
    azure_cis_scanner.set_credentials_tuples()
    """
    pass

def test_unicode_insert_error():
    """
    Cannot decode byte 0xef in ~/.azure/azureProfile.json
    Find cause of why invisible bytes are inserted.
    Fix is to open the file and delete the invisible bytes before the first '{'
    UnicodeDecodeError: 'ascii' codec can't decode byte 0xef in position 0: ordinal not in range(128)
    """
# In addition, we should use vagrant or azure deployments of the scanner to Ubuntu and Windows virtual machines
# to ensure cross-platform behavior.
    pass


import os
import yaml
import functools

# set paths depending on if we are in the container or not
if os.path.exists('/praetorian-tools/azure_cis_scanner/'):
    cis_scanner_root = '/praetorian-tools/azure_cis_scanner/'
elif os.path.exists(os.path.expanduser('~/praetorian-tools/azure_cis_scanner/')):
    cis_scanner_root = os.path.expanduser('~/praetorian-tools/azure_cis_scanner/')
else:
    cis_scanner_root = os.path.abspath(os.path.join(os.path.dirname(__file__), '../../'))

print("cis_scanner_root", cis_scanner_root)

if os.path.exists(os.path.expanduser('~/engagements/cis_test/scans')):
    scans_base = os.path.expanduser('~/engagements/cis_test/scans')
elif os.path.exists('/engagements/cis_test/scans'):
    scans_base = '/engagements/cis_test/scans'
else:
    scans_base = os.path.join(cis_scanner_root, 'scans')

@functools.lru_cache(1, typed=False)
def get_dirs(directory):
    return list(reversed(sorted([x for x in os.listdir(directory) if os.path.isdir(directory) and not x.endswith('.DS_Store')])))


accounts = {}
with open(os.path.join(scans_base, 'accounts.json'), 'r') as f:
    accounts = yaml.load(f)

APP_ROOT = os.path.dirname(os.path.abspath(__file__))
STATIC = os.path.join(APP_ROOT, 'static')

with open(os.path.expanduser(APP_ROOT + '/cis_structure.yaml'), 'r') as f:
    cis_structure = yaml.load(f)




#!/usr/bin/env python

# distutils/setuptools install script for azure_cis_scanner
import os
from setuptools import setup, find_packages

# Package info
NAME = 'azure_cis_scanner'
ROOT = os.path.dirname(__file__)
VERSION = __import__(NAME).__version__

# Requirements
requirements = []
with open(os.path.join(os.path.dirname(os.path.realpath(__file__)), 'requirements.txt')) as f:
    for r in f.readlines():
        requirements.append(r.strip())

try:
    import matplotlib
    print('matplotlib found.  Graphing supported')
except:
    print('matplotlib not installed.  Graphing not supported')

# Setup
setup(
    name=NAME,
    version=VERSION,
    description='Azure CIS Scanner for security',
    long_description=open('README.rst').read(),
    author='kesten broughton, tanner harper',
    author_email='kesten.broughton@praetorian.com',
    url='https://github.com/praetorian-inc/azure_cis_scanner',
    entry_points={
        'console_scripts': [
            'azscan = azure_cis_scanner.__main__:main',
            #'azscanRulesGenerator = azure_cis_scanner.__rules_generator__:main',
            #'azscanListAll = azure_cis_scanner.__listall__:main'
        ]
    },
    packages=[
        'azure_cis_scanner', 
        'azure_cis_scanner.modules',
        'azure_cis_scanner.report', 
    ],
    package_data={
        'azure_cis_scanner': [
            'requirements.txt',
            'cis_structure.yaml'
        ],
        'report': [
            'reports/static/*.css',
            'reports/templates/*.html',
            'reports/requirements.txt'
        ]
    },
    include_package_data=True,
    install_requires=requirements,
    license='GNU General Public License v2 (GPLv2)',
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Developers',
        'Intended Audience :: Information Technology',
        'Intended Audience :: System Administrators',
        'Natural Language :: English',
        'License :: OSI Approved :: GNU General Public License v2 (GPLv2)',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3.6'
    ],
    scripts=[
      'bin/azscan',
      'bin/azscan.bat'
    ]
)



#%load {scanner_dir}/scanner/modules/security_center.py
import yaml
import os
import json
from azure_cis_scanner import utils

import azurerm
from azure.common.client_factory import get_client_from_cli_profile, get_client_from_auth_file
from azure.mgmt.compute import ComputeManagementClient
from azure.mgmt.resource import ResourceManagementClient, SubscriptionClient
from msrestazure.azure_active_directory import MSIAuthentication

security_center_filtered_path = os.path.join(config['filtered_data_dir'], 'security_center_filtered.json')
security_center_path = os.path.join(config['raw_data_dir'], "security_center.json")
subscription_id = config['subscription_id']


def get_security_center(security_center_path):
    """
    Query Azure api for storage accounts info and save to disk
    """

    access_token = azurerm.get_access_token_from_cli()
    print("access_token", access_token)
    
    # The usual split breaks up bearer token phrase with spaces and breaks
    security_center_cmd = 'curl -X GET -H'.split() + ["Authorization: Bearer {token}".format(token=access_token)] + \
    ["-H","Content-Type: application/json"] + \
    ["https://management.azure.com/subscriptions/{subscription_id}/\providers/microsoft.Security/policies?api-version=2015-06-01-preview".format(subscription_id=subscription_id)]


    security_center = json.loads(utils.call(security_center_cmd))
    security_center = security_center['value']
        
    with open(security_center_path, 'w') as f:
        yaml.dump(security_center, f)
    return security_center

def load_security_center(security_center_filtered_path):
    with open(security_center_path, 'r') as f:
        security_center = yaml.load(f)
    return security_center

def get_data():
    """
    Generate json for the security_center findings
    """
    get_security_center(security_center_path)

def test_controls():
    """
    Generate filtered (failing) output in json
    """
    security_center = load_security_center(security_center_path)
    security_center_results = {}

    security_center_results['automatic_provisioning_of_monitoring_agent_is_set_to_on'] = automatic_provisioning_of_monitoring_agent_is_set_to_on_2_2(security_center)
    security_center_results['system_updates_is_set_to_on'] = system_updates_is_set_to_on_2_3(security_center)
    security_center_results['security_configurations_is_set_to_on'] = security_configurations_is_set_to_on_2_4(security_center)
    security_center_results['endpoint_protection_is_set_to_on'] = endpoint_protection_is_set_to_on_2_5(security_center)
    security_center_results['disk_encryption_is_set_to_on'] = disk_encryption_is_set_to_on_2_6(security_center)
    security_center_results['network_security_groups_is_set_to_on'] = network_security_groups_is_set_to_on_2_7(security_center)
    security_center_results['web_application_firewall_is_set_to_on'] = web_application_firewall_is_set_to_on_2_8(security_center)
    security_center_results['next_generation_firewall_is_set_to_on'] = next_generation_firewall_is_set_to_on_2_9(security_center)
    security_center_results['vulnerability_assessment_is_set_to_on'] = vulnerability_assessment_is_set_to_on_2_10(security_center)
    security_center_results['storage_encryption_is_set_to_on'] = storage_encryption_is_set_to_on_2_11(security_center)
    security_center_results['just_in_time_access_is_set_to_on'] = just_in_time_access_is_set_to_on_2_12(security_center)
    security_center_results['adaptive_application_controls_is_set_to_on'] = adaptive_application_controls_is_set_to_on_2_13(security_center)
    security_center_results['sql_auditing_and_threat_detection_is_set_to_on'] = sql_auditing_and_threat_detection_is_set_to_on_2_14(security_center)
    security_center_results['sql_encryption_is_set_to_on'] = sql_encryption_is_set_to_on_2_15(security_center)
    security_center_results['security_contact_emails_is_set'] = security_contact_emails_is_set_2_16(security_center)
    security_center_results['security_contact_phone_number_is_set'] = security_contact_phone_number_is_set_2_17(security_center)
    security_center_results['send_me_emails_about_alerts_is_set_to_on'] = send_me_emails_about_alerts_is_set_to_on_2_18(security_center)
    security_center_results['send_email_also_to_subscription_owners_is_set_to_on'] = send_email_also_to_subscription_owners_is_set_to_on_2_19(security_center)
                
    with open(security_center_filtered_path, 'w') as f:
        json.dump(security_center_results, f, indent=4, sort_keys=True)
    return security_center_results

def automatic_provisioning_of_monitoring_agent_is_set_to_on_2_2(security_center):
    items_flagged_list = []
    for item in security_center:
        resource_group = item['name']
        automatic_provisioning_of_monitoring_agent = item['properties']['logCollection']
        if automatic_provisioning_of_monitoring_agent != "On":
            items_flagged_list.append((resource_group))

    stats = {'items_flagged': len(items_flagged_list),
             'items_checked': len(security_center)}
    metadata = {"finding_name": "automatic_provisioning_of_monitoring_agent_is_set_to_on",
                "negative_name": "automatic_provisioning_of_monitoring_agent_not_on",
                "columns": ["Resource Group"]}
    
    return {"items": items_flagged_list, "stats": stats, "metadata": metadata}
    
def system_updates_is_set_to_on_2_3(security_center):
    items_flagged_list = []
    for item in security_center:
        resource_group = item['name']
        system_updates = item['properties']['recommendations']['patch']
        if system_updates != "On":
            items_flagged_list.append((resource_group))

    stats = {'items_flagged': len(items_flagged_list),
             'items_checked': len(security_center)}
    metadata = {"finding_name": "system_updates_is_set_to_on",
                "negative_name": "system_updates_not_on",
                "columns": ["Resource Group"]}
    
    return {"items": items_flagged_list, "stats": stats, "metadata": metadata}    

def security_configurations_is_set_to_on_2_4(security_center):
    items_flagged_list = []
    for item in security_center:
        resource_group = item['name']
        security_configurations = item['properties']['recommendations']['baseline']
        if security_configurations != "On":
            items_flagged_list.append((resource_group))
            
    stats = {'items_flagged': len(items_flagged_list),
             'items_checked': len(security_center)}
    metadata = {"finding_name": "security_configurations_is_set_to_on",
                "negative_name": "security_configurations_not_on",
                "columns": ["Resource Group"]}
    
    return {"items": items_flagged_list, "stats": stats, "metadata": metadata} 

def endpoint_protection_is_set_to_on_2_5(security_center):
    items_flagged_list = []
    for item in security_center:
        resource_group = item['name']
        endpoint_protection = item['properties']['recommendations']['antimalware']
        if endpoint_protection != "On":
            items_flagged_list.append((resource_group))

    stats = {'items_flagged': len(items_flagged_list),
             'items_checked': len(security_center)}
    metadata = {"finding_name": "endpoint_protection_is_set_to_on",
                "negative_name": "endpoint_protection_not_on",
                "columns": ["Resource Group"]}
    
    return {"items": items_flagged_list, "stats": stats, "metadata": metadata}

def disk_encryption_is_set_to_on_2_6(security_center):
    items_flagged_list = []
    for item in security_center:
        resource_group = item['name']
        disk_encryption = item['properties']['recommendations']['diskEncryption']
        if disk_encryption != "On":
            items_flagged_list.append((resource_group))

    stats = {'items_flagged': len(items_flagged_list),
             'items_checked': len(security_center)}
    metadata = {"finding_name": "disk_encryption_is_set_to_on",
                "negative_name": "disk_encryption_not_on",
                "columns": ["Resource Group"]}
    
    return {"items": items_flagged_list, "stats": stats, "metadata": metadata}

def network_security_groups_is_set_to_on_2_7(security_center):
    items_flagged_list = []
    for item in security_center:
        resource_group = item['name']
        nsgs = item['properties']['recommendations']['nsgs']
        if nsgs != "On":
            items_flagged_list.append((resource_group))

    stats = {'items_flagged': len(items_flagged_list),
             'items_checked': len(security_center)}
    metadata = {"finding_name": "network_security_groups_is_set_to_on",
                "negative_name": "network_security_groups_not_on",
                "columns": ["Resource Group"]}
    
    return {"items": items_flagged_list, "stats": stats, "metadata": metadata}

def web_application_firewall_is_set_to_on_2_8(security_center):
    items_flagged_list = []
    for item in security_center:
        resource_group = item['name']
        waf = item['properties']['recommendations']['waf']
        if waf != "On":
            items_flagged_list.append((resource_group))
    
    stats = {'items_flagged': len(items_flagged_list),
             'items_checked': len(security_center)}
    metadata = {"finding_name": "web_application_firewall_is_set_to_on",
                "negative_name": "web_application_firewall_not_on",
                "columns": ["Resource Group"]}
    
    return {"items": items_flagged_list, "stats": stats, "metadata": metadata}

def next_generation_firewall_is_set_to_on_2_9(security_center):
    items_flagged_list = []
    for item in security_center:
        resource_group = item['name']
        ngfw = item['properties']['recommendations']['ngfw']
        if ngfw != "On":
            items_flagged_list.append((resource_group))
    
    stats = {'items_flagged': len(items_flagged_list),
             'items_checked': len(security_center)}
    metadata = {"finding_name": "automatic_provisioning_of_monitoring_agent_is_set_to_on",
                "negative_name": "automatic_provisioning_of_monitoring_agent_not_on",
                "columns": ["Resource Group"]}
    
    return {"items": items_flagged_list, "stats": stats, "metadata": metadata}

def vulnerability_assessment_is_set_to_on_2_10(security_center):
    items_flagged_list = []
    for item in security_center:
        resource_group = item['name']
        vulnerability_assessment = item['properties']['recommendations']['vulnerabilityAssessment']
        if vulnerability_assessment != "On":
            items_flagged_list.append((resource_group))
            
    stats = {'items_flagged': len(items_flagged_list),
             'items_checked': len(security_center)}
    metadata = {"finding_name": "automatic_provisioning_of_monitoring_agent_is_set_to_on",
                "negative_name": "automatic_provisioning_of_monitoring_agent_not_on",
                "columns": ["Resource Group"]}
    
    return {"items": items_flagged_list, "stats": stats, "metadata": metadata}

def storage_encryption_is_set_to_on_2_11(security_center):
    items_flagged_list = []
    for item in security_center:
        resource_group = item['name']
        storage_encryption = item['properties']['recommendations']['storageEncryption']
        if storage_encryption != "On":
            items_flagged_list.append((resource_group))

    stats = {'items_flagged': len(items_flagged_list),
             'items_checked': len(security_center)}
    metadata = {"finding_name": "storage_encryption_is_set_to_on",
                "negative_name": "storage_encryption_not_on",
                "columns": ["Resource Group"]}
    
    return {"items": items_flagged_list, "stats": stats, "metadata": metadata}

def just_in_time_access_is_set_to_on_2_12(security_center):
    items_flagged_list = []
    for item in security_center:
        resource_group = item['name']
        jit = item['properties']['recommendations']['jitNetworkAccess']
        if jit != "On":
            items_flagged_list.append((resource_group))

    stats = {'items_flagged': len(items_flagged_list),
             'items_checked': len(security_center)}
    metadata = {"finding_name": "just_in_time_access_is_set_to_on",
                "negative_name": "just_in_time_access_not_on",
                "columns": ["Resource Group"]}
    
    return {"items": items_flagged_list, "stats": stats, "metadata": metadata}

def adaptive_application_controls_is_set_to_on_2_13(security_center):
    items_flagged_list = []
    for item in security_center:
        resource_group = item['name']
        security_configurations = item['properties']['recommendations']['appWhitelisting']
        if security_configurations != "On":
            items_flagged_list.append((resource_group))

    stats = {'items_flagged': len(items_flagged_list),
             'items_checked': len(security_center)}
    metadata = {"finding_name": "adaptive_application_controls_is_set_to_on",
                "negative_name": "adaptive_application_controls_noto_on",
                "columns": ["Resource Group"]}
    
    return {"items": items_flagged_list, "stats": stats, "metadata": metadata}

def sql_auditing_and_threat_detection_is_set_to_on_2_14(security_center):
    items_flagged_list = []
    for item in security_center:
        resource_group = item['name']
        sqlAuditing = item['properties']['recommendations']['sqlAuditing']
        if sqlAuditing != "On":
            items_flagged_list.append((resource_group))

    stats = {'items_flagged': len(items_flagged_list),
             'items_checked': len(security_center)}
    metadata = {"finding_name": "sql_auditing_and_threat_detection_is_set_to_on",
                "negative_name": "sql_auditing_and_threat_detection_not_on",
                "columns": ["Resource Group"]}
    
    return {"items": items_flagged_list, "stats": stats, "metadata": metadata}

def sql_encryption_is_set_to_on_2_15(security_center):
    items_flagged_list = []
    for item in security_center:
        resource_group = item['name']
        sql_tde = item['properties']['recommendations']['sqlTde']
        if sql_tde != "On":
            items_flagged_list.append((resource_group))

    stats = {'items_flagged': len(items_flagged_list),
             'items_checked': len(security_center)}
    metadata = {"finding_name": "sql_encryption_is_set_to_on",
                "negative_name": "sql_encryption_not_on",
                "columns": ["Resource Group"]}
    
    return {"items": items_flagged_list, "stats": stats, "metadata": metadata}

def security_contact_emails_is_set_2_16(security_center):
    items_flagged_list = []
    for item in security_center:
        resource_group = item['name']
        emails = item['properties']['securityContactConfiguration']['securityContactEmails']
        if not emails:
            items_flagged_list.append((resource_group))

    stats = {'items_flagged': len(items_flagged_list),
             'items_checked': len(security_center)}
    metadata = {"finding_name": "security_contact_emails_is_set",
                "negative_name": "security_contact_emails_not_set",
                "columns": ["Resource Group"]}
    
    return {"items": items_flagged_list, "stats": stats, "metadata": metadata}

def security_contact_phone_number_is_set_2_17(security_center):
    items_flagged_list = []
    for item in security_center:
        resource_group = item['name']
        phone = item['properties']['securityContactConfiguration']['securityContactPhone']
        if not phone:
            items_flagged_list.append((resource_group))
    stats = {'items_flagged': len(items_flagged_list),
             'items_checked': len(security_center)}
    metadata = {"finding_name": "security_contact_phone_number_is_set",
                "negative_name": "security_contact_phone_number_not_set",
                "columns": ["Resource Group"]}
    
    return {"items": items_flagged_list, "stats": stats, "metadata": metadata}

def send_me_emails_about_alerts_is_set_to_on_2_18(security_center):
    items_flagged_list = []
    for item in security_center:
        resource_group = item['name']
        notifications = item['properties']['securityContactConfiguration']['areNotificationsOn']
        if notifications != "On":
            items_flagged_list.append((resource_group))

    stats = {'items_flagged': len(items_flagged_list),
             'items_checked': len(security_center)}
    metadata = {"finding_name": "send_email_alerts_about_alerts_is_set_to_on",
                "negative_name": "send_email_alerts_about_alerts_not_on",
                "columns": ["Resource Group"]}
    
    return {"items": items_flagged_list, "stats": stats, "metadata": metadata}

def send_email_also_to_subscription_owners_is_set_to_on_2_19(security_center):
    items_flagged_list = []
    for item in security_center:
        resource_group = item['name']
        send_admin = item['properties']['securityContactConfiguration']['sendToAdminOn']
        if send_admin != "On":
            items_flagged_list.append((resource_group))

    stats = {'items_flagged': len(items_flagged_list),
             'items_checked': len(security_center)}
    metadata = {"finding_name": "send_email_also_to_subscription_owners_is_set_to_on",
                "negative_name": "send_email_also_to_subscription_owners_is_set_to_on",
                "columns": ["Resource Group"]}
    
    return {"items": items_flagged_list, "stats": stats, "metadata": metadata}


import functools
import yaml
from settings import *
import re
import pprint 

@functools.lru_cache(1, typed=False)
def get_dirs(directory):
    return list(reversed(sorted([x for x in os.listdir(directory) if os.path.isdir(directory) and not x.endswith('.DS_Store')])))

@functools.lru_cache(maxsize=32, typed=False)
def get_filtered_data_path(scans_data_dir, date=None):
    """
    Get the filtered data root for the scan run on date=date or latest if date=None
    Returns path, date where date is the most recent date with data <= requested date
    
    Directory structure is
    |-------------- scans_data_dir --------|
    scans_dir/<active_subscription_dirname>/<date>/<section_lowercase_underscores>.json
    """
    if date:
        if os.path.exists:
            return os.path.join(scans_data_dir, date, 'filtered'), date
        else:
            raise ValueError("Filtered data requested for {} but file does not exist at {}".format(
                date, scans_data_dir))
    else:
        dir_list = get_dirs(scans_data_dir)
        if len(dir_list) == 0:
            print("No data found in {}.  Please run scanner first".format(scans_data_dir))
        else:
            date = dir_list[0]
            return os.path.join(scans_data_dir, date, 'filtered'), date

@functools.lru_cache(maxsize=32, typed=False)
def get_filtered_data(scans_data_dir, date=None):
    """
    Returns a dict of filtered data for a specific date or latest (default)
    
    If a section is missing it will not be returned.
    The structure is {"Identity and Access Management": {"finding1": results_dict1}, "finding2": results_dict2}
    where results_dict has keys stats, metadata, items, date - where date is actual date where data was found
    """
    filtered_data_root, date = get_filtered_data_path(scans_data_dir, date)
    filtered_data = {}
    for item in cis_structure['section_ordering']:
        item = '_'.join(map(str.lower, item.split(' '))) + '_filtered.json'
        with open(item, 'r') as f:
            data = yaml.load(f)
            data['date'] = date
        filtered_data[item] = data
    return filtered_data

@functools.lru_cache(maxsize=128, typed=False)
def get_filtered_data_by_name(scans_data_dir, section_name, date=None):
    """
    Get the latest data for a section returning first found <= date
    @params sectoin_name: Name of CIS section as a string e.g. ("Identity and Access Management")
    @params date: date in format 'YYYY-M-D', i.e. strftime("%Y-%m-%d")
    @returns filtered data, date
    """
    # get date folders, most to least recent
    dir_list = get_dirs(scans_data_dir)
    section_name_file = '_'.join(map(str.lower, section_name.split(' '))) + '_filtered.json'
    for dir_date in dir_list:
        if date and (dir_date > date):
            continue
        filtered_data_path = os.path.join(scans_data_dir, dir_date, 'filtered', section_name_file)
        if os.path.exists(filtered_data_path):
            print('get_filtered_data_by_name', filtered_data_path)
            with open(filtered_data_path, 'r') as f:
                data = yaml.load(f)
                data['date'] = dir_date
                return data
    else:
        return None


@functools.lru_cache(maxsize=1, typed=False)
def get_latest_filtered_data(scans_data_dir, date=None):
    """
    Returns a dict as in get_filtered_data, but if a section is missing, it will search
    back in time for a date where the section does exist.
    """
    data = get_filtered_data(scans_data_dir, date)
    if not data:
        return None
    else:
        for section_name in cis_structure['section_ordering']:
            #section_name = '_'.join(map(lower, section_name.split(' '))) + '.json'
            if section_name not in data:
                section_data = get_filtered_data_by_name(scans_data_dir, section_name, date)
                if section_data:
                    data[section_name] = section_data
    return data

@functools.lru_cache(maxsize=1, typed=False)
def get_stats(scans_data_dir):
    stats = {}
    dir_list = get_dirs(scans_data_dir)
    print("get_stats dir_list {}".format(dir_list))
    for section_name in cis_structure['section_ordering']:
        stats[section_name] = {}
        section_name_file = '_'.join(map(str.lower, section_name.split(' '))) + '_filtered.json'
        for dir_date in dir_list:
            filtered_data_path = os.path.join(scans_data_dir, dir_date, 'filtered', section_name_file)
            if os.path.exists(filtered_data_path):
                with open(filtered_data_path, 'r') as f:
                    data = yaml.load(f)
                for finding_name, finding_data in data.items():
                    if not finding_name in stats[section_name]:
                        stats[section_name][finding_name] = {}
                    stats[section_name][finding_name][dir_date] = finding_data['stats']
    return stats

@functools.lru_cache(maxsize=1, typed=False)
def get_latest_stats(scans_data_dir):
    latest_stats = {}
    stats = get_stats(scans_data_dir)
    pprint.pprint("get_latest_stats:get_stats: {}".format(stats))
    for section_name in stats:
        latest_stats[section_name] = {}
        for finding_name in stats[section_name]:
            date = max(stats[section_name][finding_name])
            latest_stats[section_name][finding_name] = {"date": date, **stats[section_name][finding_name][date]}

    print("get_latest_stats:latest_stats: {}".format(latest_stats))
    return latest_stats

def get_finding_name(finding_name, subsection_name):
    """
    Get finding name from CIS_TOC using over-ride (finding_name) but defaulting to parsed subsection_name
    """
    if finding_name:
        return finding_name
    else:
        return underscore_name(subsection_name)

def underscore_name(subsection_name):
    return '_'.join(map(lambda x: x.lower(), subsection_name.split(' ')))

def title_except(string):
    articles = ['a', 'an', 'of', 'the', 'is', 'not', 'for', 'if']
    word_list = re.split(' ', string)       # re.split behaves as expected
    final = [word_list[0].capitalize()]
    for word in word_list[1:]:
        final.append(word if word in articles else word.capitalize())
    return " ".join(final)

def get_finding_index(findings_list, finding):
    for finding_entry in findings_list:
        if finding_entry['subsection_name'] == finding:
            return finding_entry
    raise ValueError("finding {} not found in {}".format(finding, findings_list))


# def get_summary_stats(section=None):
# 	"""
# 	Return the sum over impacted_items in all findings in a section

# 	If section == None, sum over sections
# 	"""
# 	stats = get_stats()
# 	if section:

# 	else:
# 		for section in stats:



